#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2022 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .core import LazyLoad, Iterable, XMLRemoteModel
from .. import serializers, errors, utils
from ..compat import six, Enum
from .volumes import Volume
from .cache import cache_parent


class VolumeFile(serializers.XMLSerializableModel):
    _root = 'VolumeFileModel'

    name = serializers.XMLNodeField('Name')

    @property
    def partition(self):
        return self.parent.parent

    @property
    def project(self):
        return self.partition.project

    @property
    def volume(self):
        return self.partition.volume

    def open_reader(self, **kw):
        return self.partition.open_reader(self.name, **kw)

    @property
    def path(self):
        return '/'.join((self.project.name, 'volumes', self.volume.name, self.partition.name, self.name))


class VolumePartitionMeta(XMLRemoteModel):
    length = serializers.XMLNodeField('Length', parse_callback=int, set_to_parent=True)
    file_number = serializers.XMLNodeField('FileNumber', parse_callback=int, set_to_parent=True)

    def load(self):
        url = self._parent.resource()
        params = {}

        schema_name = self._parent._get_schema_name()
        if schema_name is not None:
            params["curr_schema"] = schema_name

        resp = self._client.get(url, action='meta', params=params)
        self.parse(self._client, resp, obj=self)


class VolumePartition(LazyLoad):
    """
    Represents a partition in a volume.
    """
    __slots__ = '_volume_tunnel', '_id_thread_local', \
                'length', 'file_number'  # meta

    class Type(Enum):
        NEW = 'NEW'
        OLD = 'OLD'

    _root = 'Meta'
    _type_indicator = 'type'

    name = serializers.XMLNodeField('Name')
    owner = serializers.XMLNodeField('Owner')
    type = serializers.XMLNodeField('Type', parse_callback=lambda t: VolumePartition.Type(t.upper()))
    comment = serializers.XMLNodeField('Comment')
    creation_time = serializers.XMLNodeField('CreationTime', parse_callback=utils.parse_rfc822)
    last_modified_time = serializers.XMLNodeField('LastModifiedTime', parse_callback=utils.parse_rfc822)
    meta = serializers.XMLNodeReferenceField(VolumePartitionMeta, 'Meta')

    _download_id = utils.thread_local_attribute('_id_thread_local', lambda: None)
    _upload_id = utils.thread_local_attribute('_id_thread_local', lambda: None)

    def __init__(self, **kwargs):
        super(VolumePartition, self).__init__(**kwargs)
        self._volume_tunnel = None
        self.meta = VolumePartitionMeta(parent=self, client=self._client)

        try:
            del self._id_thread_local
        except AttributeError:
            pass

    def reload(self):
        resp = self._client.get(self.resource(), curr_schema=self._get_schema_name())
        self.parse(self._client, resp, obj=self)
        self.meta.load()
        self._loaded = True

    @property
    def volume(self):
        return self.parent

    def __getattribute__(self, attr):
        val = object.__getattribute__(self, attr)
        if val is None and attr in getattr(VolumePartitionMeta, '__fields'):
            self.meta.load()
            return object.__getattribute__(self, attr)

        return super(VolumePartition, self).__getattribute__(attr)

    class VolumeFiles(serializers.XMLSerializableModel):
        _root = 'Items'
        skip_null = False

        marker = serializers.XMLNodeField('Marker')
        files = serializers.XMLNodesReferencesField(VolumeFile, 'Item')
        max_items = serializers.XMLNodeField('MaxItems', parse_callback=int)

        def __getitem__(self, item):
            for f in self.iterate(name=item):
                if f.name == item:
                    return f
            raise KeyError

        def get(self, item, default=None):
            try:
                return self[item]
            except KeyError:
                return default

        def __iter__(self):
            return self.iterate()

        def iterate(self, name=None, max_items=None):
            """
            :param name: the prefix of volume name name
            :return:
            """
            params = {'expectmarker': 'true', 'path': ''}
            if name is not None:
                params['name'] = name
            if max_items is not None:
                params['maxitems'] = max_items

            schema_name = self.parent._get_schema_name()
            if schema_name is not None:
                params["curr_schema"] = schema_name

            def _it():
                last_marker = params.get('marker')
                if 'marker' in params and \
                        (last_marker is None or len(last_marker) == 0):
                    return

                url = self.parent.resource()
                resp = self.parent._client.get(url, params=params)

                v = self.parse(resp, obj=self, parent=self.parent)
                params['marker'] = v.marker

                return v.files

            while True:
                files = _it()
                if files is None:
                    break
                for f in files:
                    yield f

    @property
    def files(self):
        return self.VolumeFiles(_parent=self)

    def list_files(self):
        """
        List files in the partition.

        :return: files
        :rtype: list
        """
        return self.files.iterate()

    def _create_volume_tunnel(self, endpoint=None, quota_name=None):
        if self._volume_tunnel is not None:
            return self._volume_tunnel

        from ..tunnel import VolumeTunnel

        self._volume_tunnel = VolumeTunnel(
            client=self._client,
            project=self.project,
            endpoint=endpoint or self.project._tunnel_endpoint,
            quota_name=quota_name,
        )
        return self._volume_tunnel

    def open_reader(
        self, file_name, reopen=False, endpoint=None, start=None, length=None, quota_name=None, **kwargs
    ):
        """
        Open a volume file for read. A file-like object will be returned which can be used to read contents from
        volume files.

        :param str file_name: name of the file
        :param bool reopen: whether we need to open an existing read session
        :param str endpoint: tunnel service URL
        :param start: start position
        :param length: length limit
        :param quota_name: name of tunnel quota
        :param compress_option: the compression algorithm, level and strategy
        :type compress_option: :class:`odps.tunnel.CompressOption`

        :Example:
        >>> with partition.open_reader('file') as reader:
        >>>     [print(line) for line in reader]
        """
        tunnel = self._create_volume_tunnel(endpoint=endpoint, quota_name=quota_name)
        download_id = self._download_id if reopen else None
        download_session = tunnel.create_download_session(
            volume=self.volume.name, partition_spec=self.name,
            file_name=file_name, download_id=download_id, **kwargs
        )
        self._download_id = download_session.id

        open_args = {}
        if start is not None:
            open_args['start'] = start
        if length is not None:
            open_args['length'] = length
        return download_session.open(**open_args)

    def open_writer(self, reopen=False, endpoint=None, quota_name=None, **kwargs):
        """
        Open a volume partition to write to. You can use `open` method to open a file inside the volume and write to it,
        or use `write` method to write to specific files.

        :param bool reopen: whether we need to open an existing write session
        :param str endpoint: tunnel service URL
        :param quota_name: name of tunnel quota
        :param compress_option: the compression algorithm, level and strategy
        :type compress_option: :class:`odps.tunnel.CompressOption`
        :Example:
        >>> with partition.open_writer() as writer:
        >>>     writer.open('file1').write('some content')
        >>>     writer.write('file2', 'some content')
        """
        tunnel = self._create_volume_tunnel(endpoint=endpoint, quota_name=quota_name)
        upload_id = self._upload_id if reopen else None
        upload_session = tunnel.create_upload_session(
            volume=self.volume.name, partition_spec=self.name,
            upload_id=upload_id, **kwargs
        )
        self._upload_id = upload_session.id
        file_dict = dict()

        class FilesWriter(object):
            @property
            def status(self):
                return upload_session.status

            @staticmethod
            def open(file_name, **kwargs):
                if file_name in file_dict:
                    return file_dict[file_name]
                writer = upload_session.open(file_name, **kwargs)
                file_dict[file_name] = writer
                return writer

            @staticmethod
            def write(file_name, buf, **kwargs):
                writer = FilesWriter.open(file_name, **kwargs)
                writer.write(buf)

            @staticmethod
            def close():
                for w in six.itervalues(file_dict):
                    w.close()

                upload_session.commit(list(six.iterkeys(file_dict)))

            def __enter__(self):
                return self

            def __exit__(self, exc_type, exc_val, exc_tb):
                self.close()

        return FilesWriter()


@cache_parent
class PartedVolume(Volume):
    """
    PartedVolume represents the old-fashioned partitioned volume in ODPS.
    """
    class Partitions(Iterable):
        _root = 'Volume'

        marker = serializers.XMLNodeField('Marker')
        partitions = serializers.XMLNodesReferencesField(
            VolumePartition, 'Partitions', 'Partition'
        )
        max_items = serializers.XMLNodeField('MaxItems', parse_callback=int)

        def _get(self, item):
            return VolumePartition(client=self._client, parent=self.parent, name=item)

        def __contains__(self, item):
            if isinstance(item, six.string_types):
                partition = self._get(item)
            elif isinstance(item, VolumePartition):
                partition = item
            else:
                return False

            try:
                partition.reload()
                return True
            except errors.NoSuchObject:
                return False

        def __iter__(self):
            return self.iterate()

        def iterate(self, name=None, owner=None):
            """
            :param name: the prefix of volume name name
            :param owner:
            :return:
            """
            params = {'expectmarker': 'true'}
            if name is not None:
                params['name'] = name
            if owner is not None:
                params['owner'] = owner

            schema_name = self.parent._get_schema_name()
            if schema_name is not None:
                params["curr_schema"] = schema_name

            def _it():
                last_marker = params.get('marker')
                if 'marker' in params and \
                        (last_marker is None or len(last_marker) == 0):
                    return

                url = self.parent.resource()
                resp = self._client.get(url, params=params)

                v = PartedVolume.Partitions.parse(self._client, resp, obj=self)
                params['marker'] = v.marker

                return v.partitions

            while True:
                partitions = _it()
                if partitions is None:
                    break
                for partition in partitions:
                    yield partition

        def delete(self, name):
            if not isinstance(name, Volume):
                partition = self._get(name)
            else:
                partition = name

            url = partition.resource()
            del self[name]
            self._client.delete(url, curr_schema=partition._get_schema_name())

    @property
    def partitions(self):
        return PartedVolume.Partitions(parent=self, client=self._client)

    def list_partitions(self, owner=None):
        """
        List partitions.

        :return: partitions
        :rtype: list
        """
        return self.partitions.iterate(owner=owner)

    def exist_partition(self, name):
        """
        If the volume with given name exists in a partition or not.

        :param str name: partition name
        """
        return name in self.partitions

    def get_partition(self, name):
        """
        Get partition from volume by name

        :param str name: partition name
        :return: partition
        :rtype: :class:`odps.models.VolumePartition`
        """
        return self.partitions[name]

    def delete_partition(self, name):
        """
        Delete partition by given name

        :param str name: partition name
        """
        return self.partitions.delete(name)

    def open_reader(self, partition, file_name, endpoint=None, start=None, length=None, **kwargs):
        """
        Open a volume file for read. A file-like object will be returned which can be used to read contents from
        volume files.

        :param str partition: name of the partition
        :param str file_name: name of the file
        :param str endpoint: tunnel service URL
        :param start: start position
        :param length: length limit
        :param compress_option: the compression algorithm, level and strategy
        :type compress_option: :class:`odps.tunnel.CompressOption`

        :Example:
        >>> with volume.open_reader('part', 'file') as reader:
        >>>     [print(line) for line in reader]
        """
        return self.partitions[partition].open_reader(file_name, endpoint=endpoint, start=start, length=length,
                                                      **kwargs)

    def open_writer(self, partition, endpoint=None, **kwargs):
        """
        Open a volume partition to write to. You can use `open` method to open a file inside the volume and write to it,
        or use `write` method to write to specific files.

        :param str partition: name of the partition
        :param str endpoint: tunnel service URL
        :param compress_option: the compression algorithm, level and strategy
        :type compress_option: :class:`odps.tunnel.CompressOption`
        :Example:
        >>> with volume.open_writer() as writer:
        >>>     writer.open('file1').write('some content')
        >>>     writer.write('file2', 'some content')
        """
        return self.partitions[partition].open_writer(endpoint=endpoint, **kwargs)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2022 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json

from ..core import Iterable, LazyLoad
from ...compat import six
from ... import serializers, errors


class Role(LazyLoad):
    __slots__ = '_policy_cache',
    name = serializers.XMLNodeField('Name')
    comment = serializers.XMLNodeField('Comment')

    def __init__(self, **kw):
        super(Role, self).__init__(**kw)
        self._policy_cache = None

    def reload(self):
        self._policy_cache = None
        resp = self._client.get(self.resource())
        self.parse(self._client, resp, obj=self)

    @property
    def users(self):
        from .users import Users
        params = dict(users='')
        resp = self._client.get(self.resource(), params=params)
        users = Users.parse(self._client, resp, parent=self.project)
        users._iter_local = True
        return users

    @property
    def policy(self):
        if self._policy_cache is None:
            params = dict(policy='')
            resp = self._client.get(self.resource(), params=params)
            self._policy_cache = resp.content.decode() if six.PY3 else resp.content
        return json.loads(self._policy_cache)

    @policy.setter
    def policy(self, value):
        if isinstance(value, (dict, list)):
            value = json.dumps(value)
        self._policy_cache = value
        params = dict(policy='')
        self._client.put(self.resource(), data=value, params=params)

    def grant_to(self, name):
        from .users import User
        if isinstance(name, User):
            name = name.display_name
        self.project.run_security_query('grant %s to %s' % (self.name, name))

    def revoke_from(self, name):
        from .users import User
        if isinstance(name, User):
            name = name.display_name
        self.project.run_security_query('revoke %s from %s' % (self.name, name))


class Roles(Iterable):
    __slots__ = '_iter_local',
    roles = serializers.XMLNodesReferencesField(Role, 'Role')

    def __init__(self, **kw):
        self._iter_local = False
        super(Roles, self).__init__(**kw)

    def _get(self, item):
        return Role(client=self._client, parent=self, name=item)

    def __contains__(self, item):
        if isinstance(item, six.string_types):
            role = self._get(item)
        elif isinstance(item, Role):
            role = item
        else:
            return False

        if not self._iter_local:
            try:
                role.reload()
                return True
            except errors.NoSuchObject:
                return False
        else:
            return any(r.name == role.name for r in self.roles)

    def __iter__(self):
        return self.iterate()

    @property
    def project(self):
        return self.parent

    def create(self, name):
        self.project.run_security_query('create role %s' % name)
        return Role(client=self._client, parent=self, name=name)

    def iterate(self, name=None):
        """
        :return:
        """
        if not self._iter_local:
            params = dict()
            if name is not None:
                params['name'] = name

            url = self.resource()
            resp = self._client.get(url, params=params)

            Roles.parse(self._client, resp, obj=self)

        for role in self.roles:
            yield role

    def delete(self, name):
        if isinstance(name, Role):
            name = name.name

        del self[name]  # delete from cache
        self.project.run_security_query('drop role %s' % name)



# -*- coding: utf-8 -*-
# Copyright 1999-2022 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import calendar
import json
import logging
import operator
from datetime import datetime

from requests import ConnectTimeout as RequestsConnectTimeout

from . import utils
from .compat import (
    six,
    reduce,
    ElementTree as ET,
    ElementTreeParseError as ETParseError,
    TimeoutError
)

logger = logging.getLogger(__name__)


class DatetimeOverflowError(OverflowError):
    pass


class DependencyNotInstalledError(Exception):
    pass


class InteractiveError(Exception):
    pass


def parse_response(resp, endpoint=None, tag=None):
    """Parses the content of response and returns an exception object.
    """
    try:
        try:
            content = resp.content
            root = ET.fromstring(content)
            code = root.find('./Code').text
            msg = root.find('./Message').text
            request_id = root.find('./RequestId').text
            host_id = root.find('./HostId').text
        except ETParseError:
            request_id = resp.headers.get('x-odps-request-id', None)
            if len(resp.content) > 0:
                obj = json.loads(resp.text)
                msg = obj['Message']
                code = obj.get('Code')
                host_id = obj.get('HostId')
                if request_id is None:
                    request_id = obj.get('RequestId')
            else:
                raise
        clz = globals().get(code, ODPSError)
        return clz(
            msg, request_id=request_id, code=code, host_id=host_id, endpoint=endpoint, tag=tag
        )
    except:
        # Error occurred during parsing the response. We ignore it and delegate
        # the situation to caller to handle.
        logger.debug(utils.stringify_expt())

    if resp.status_code == 404:
        return NoSuchObject('No such object.', endpoint=endpoint, tag=tag)
    elif resp.status_code == 401:
        return Unauthorized('Unauthorized.', endpoint=endpoint, tag=tag)
    else:
        text = resp.content.decode() if six.PY3 else resp.content
        if text:
            if resp.status_code == 502 and _nginx_bad_gateway_message in text:
                return BadGatewayError(
                    text, code=str(resp.status_code), endpoint=endpoint, tag=tag
                )
            else:
                return ODPSError(text, code=str(resp.status_code), endpoint=endpoint, tag=tag)
        else:
            return ODPSError(str(resp.status_code), endpoint=endpoint, tag=tag)


def throw_if_parsable(resp, endpoint=None, tag=None):
    """Try to parse the content of the response and raise an exception
    if necessary.
    """
    raise parse_response(resp, endpoint, tag)


_CODE_MAPPING = {
    "ODPS-0010000": "InternalServerError",
    "ODPS-0110141": "DataVersionError",
    "ODPS-0123055": "ScriptError",
    "ODPS-0130131": "NoSuchTable",
    "ODPS-0130013": "NoPermission",
    "ODPS-0130161": "ParseError",
    "ODPS-0430055": "InternalConnectionError",
}

_SQA_CODE_MAPPING = {
    'ODPS-180': 'SQAGenericError',
    'ODPS-181': 'SQARetryError',
    'ODPS-182': 'SQAAccessDenied',
    'ODPS-183': 'SQAResourceNotEnough',
    'ODPS-184': 'SQAServiceUnavailable',
    'ODPS-185': 'SQAUnsupportedFeature',
    'ODPS-186': 'SQAQueryTimedout',
}

_nginx_bad_gateway_message = "the page you are looking for is currently unavailable"


def parse_instance_error(msg):
    msg = utils.to_str(msg)
    msg_parts = reduce(operator.add, (pt.split(':') for pt in msg.split(' - ')))
    msg_parts = [pt.strip() for pt in msg_parts]
    try:
        msg_code = next(p for p in msg_parts if p.startswith('ODPS-'))
        if msg_code in _CODE_MAPPING:
            cls = globals().get(_CODE_MAPPING[msg_code], ODPSError)
        elif len(msg_code) > 8 and msg_code[:8] in _SQA_CODE_MAPPING:
            # sometimes SQA will report nested odps errors.
            # return the outer error type instead of the inner one.
            cls = globals().get(_SQA_CODE_MAPPING[msg_code[:8]], ODPSError)
            return cls(msg, code=msg_code)
        else:
            cls = ODPSError
    except StopIteration:
        cls = ODPSError
        msg_code = None

    return cls(msg, code=msg_code)


class ODPSError(RuntimeError):
    """Base class of ODPS error"""
    def __init__(
        self, msg, request_id=None, code=None, host_id=None, instance_id=None, endpoint=None, tag=None
    ):
        super(ODPSError, self).__init__(msg)
        self.request_id = request_id
        self.instance_id = instance_id
        self.code = code
        self.host_id = host_id
        self.endpoint = endpoint
        self.tag = tag

    def __str__(self):
        message = self.args[0]

        head_parts = []
        if self.code:
            head_parts.append("%s:" % self.code)
        if self.request_id:
            head_parts.append("RequestId: %s" % self.request_id)
        if self.instance_id:
            head_parts.append("InstanceId: %s" % self.instance_id)
        if self.tag:
            head_parts.append("Tag: %s" % self.tag)
        if self.endpoint:
            head_parts.append("Endpoint: %s" % self.endpoint)

        if head_parts:
            return '%s\n%s' % (" ".join(head_parts), message)
        return message

    @classmethod
    def parse(cls, resp):
        return parse_response(resp)


class ODPSClientError(ODPSError):
    pass


class ConnectTimeout(ODPSError, TimeoutError, RequestsConnectTimeout):
    pass


class DataHealthManagerError(ODPSError):
    pass


class ServerDefinedException(ODPSError):
    pass


# A long list of server defined exceptions

class MethodNotAllowed(ServerDefinedException):
    pass


class NoSuchObject(ServerDefinedException):
    pass


class NoSuchPartition(NoSuchObject):
    pass


class NoSuchPath(NoSuchObject):
    pass


class NoSuchTable(NoSuchObject):
    pass


class InvalidArgument(ServerDefinedException):
    pass


class AuthorizationRequired(ServerDefinedException):
    pass


class Unauthorized(AuthorizationRequired):
    pass


class SchemaParseError(ServerDefinedException):
    pass


class InvalidStateSetting(ServerDefinedException):
    pass


class InvalidProjectTable(ServerDefinedException):
    pass


class NoPermission(ServerDefinedException):
    pass


class InternalServerError(ServerDefinedException):
    pass


class ReadMetaError(InternalServerError):
    pass


class ServiceUnavailable(InternalServerError):
    pass


class ScriptError(ServerDefinedException):
    pass


class ParseError(ServerDefinedException):
    def __init__(self, *args, **kw):
        super(ParseError, self).__init__(*args, **kw)
        self.statement = None

    def __str__(self):
        message = super(ParseError, self).__str__()
        if self.statement is None:
            return message
        first_row, rests = message.split("\n", 1)
        statement_row = "SQL Statement: " + self.statement
        return "\n".join([first_row, statement_row, rests])


class DataVersionError(InternalServerError):
    pass


class BadGatewayError(InternalServerError):
    pass


class InstanceTypeNotSupported(ServerDefinedException):
    pass


class InvalidParameter(ServerDefinedException):
    pass


class StreamSessionNotFound(ServerDefinedException):
    pass


class UpsertSessionNotFound(ServerDefinedException):
    pass


class OverwriteModeNotAllowed(ServerDefinedException):
    pass


class TableModified(ServerDefinedException):
    pass


class RequestTimeTooSkewed(ServerDefinedException):
    def __init__(self, msg, *args, **kwargs):
        super(RequestTimeTooSkewed, self).__init__(msg, *args, **kwargs)
        try:
            parts = msg.split(',')
            kv_dict = dict(tuple(s.strip() for s in p.split(':', 1)) for p in parts)
            self.max_interval_date = int(kv_dict['max_interval_date'])
            self.expire_date = self._parse_error_date(kv_dict['expire_date'])
            self.now_date = self._parse_error_date(kv_dict['now_date'])
        except:
            self.max_interval_date = None
            self.expire_date = None
            self.now_date = None

    @staticmethod
    def _parse_error_date(date_str):
        date_obj = datetime.strptime(date_str, '%Y-%m-%dT%H:%M:%S.%fZ')
        micros = date_obj.microsecond
        return datetime.fromtimestamp(calendar.timegm(date_obj.timetuple())).replace(microsecond=micros)

# Handling error code typo in ODPS error message
RequestTimeTooSkewd = RequestTimeTooSkewed


class NotSupportedError(ODPSError):
    pass


class WaitTimeoutError(ODPSError, TimeoutError):
    pass


class SecurityQueryError(ODPSError):
    pass


class NoSuchProject(ODPSError):
    pass


class OSSSignUrlError(ODPSError):
    def __init__(self, err):
        if isinstance(err, six.string_types):
            super(OSSSignUrlError, self).__init__(err)
            self.oss_exception = None
        else:
            super(OSSSignUrlError, self).__init__(str(err))
            self.oss_exception = err


class SQAError(ODPSError):
    pass


class SQAGenericError(SQAError):
    pass


# if this error is thrown, you may retry your request.
class SQARetryError(SQAError):
    pass


class SQAAccessDenied(SQAError):
    pass


class SQAResourceNotEnough(SQAError):
    pass


class SQAServiceUnavailable(SQAError):
    pass


class SQAUnsupportedFeature(SQAError):
    pass


class SQAQueryTimedout(SQAError):
    pass



# -*- coding: utf-8 -*-
# Copyright 1999-2022 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from libc.stdint cimport *
from libc.string cimport *
from libcpp.string cimport string

from . import errno
from ..pb.wire_format import TAG_TYPE_BITS as PY_TAG_TYPE_BITS, _TAG_TYPE_MASK as _PY_TAG_TYPE_MASK

cdef:
    int TAG_TYPE_BITS = PY_TAG_TYPE_BITS
    int _TAG_TYPE_MASK = _PY_TAG_TYPE_MASK

    int BD_SUCCESS = errno.BD_SUCCESS
    int BD_BUFFER_EXHAUSTED = errno.BD_BUFFER_EXHAUSTED
    int BD_CHECKSUM_INVALID = errno.BD_CHECKSUM_INVALID
    int BD_COUNT_NOT_MATCH = errno.BD_COUNT_NOT_MATCH
    int BD_INVALID_STREAM_DATA = errno.BD_INVALID_STREAM_DATA
    int BD_INVALID_PB_TAG = errno.BD_INVALID_PB_TAG


cdef class Decoder:
    cdef init(self, char *buf_ptr, int buf_len):
        self._buf_ptr = buf_ptr
        self._buf_len = buf_len
        self._last_error = BD_SUCCESS
        self._pos = 0

    cdef int get_last_error(self) nogil:
        return self._last_error

    cdef void set_last_error(self, int errno) nogil:
        self._last_error = errno

    cdef int position(self) nogil:
        return self._pos

    cdef void add_offset(self, int n) nogil:
        self._pos += n

    cdef int32_t read_field_number(self) nogil:
        cdef int32_t tag_and_type
        tag_and_type = self.read_uint32()
        return tag_and_type >> TAG_TYPE_BITS

    cdef FieldParam read_field_number_and_wire_type(self) nogil:
        cdef int32_t tag_and_type
        cdef FieldParam fp
        tag_and_type = self.read_uint32()
        fp.field_number = tag_and_type >> TAG_TYPE_BITS
        fp.wire_type = tag_and_type & _TAG_TYPE_MASK
        return fp

    cdef int32_t read_sint32(self) nogil:
        return self._get_signed_varint32()

    cdef uint32_t read_uint32(self) nogil:
        return self._get_varint32()

    cdef int64_t read_sint64(self) nogil:
        return self._get_signed_varint64()

    cdef uint64_t read_uint64(self) nogil:
        return self._get_varint64()

    cdef bint read_bool(self) nogil:
        return self._get_varint32()

    cdef double read_double(self) nogil:
        cdef double retval

        if sizeof(double) + self._pos > self._buf_len:
            self._last_error = BD_BUFFER_EXHAUSTED
            return 0.0

        memcpy(&retval, self._buf_ptr + self._pos, sizeof(double))
        self._pos += sizeof(double)
        return retval

    cdef float read_float(self) nogil:
        cdef float retval

        if sizeof(float) + self._pos > self._buf_len:
            self._last_error = BD_BUFFER_EXHAUSTED
            return 0.0

        memcpy(&retval, self._buf_ptr + self._pos, sizeof(float))
        self._pos += sizeof(float)
        return retval

    cdef string read_string(self) nogil:
        cdef int size
        cdef int old_pos = self._pos

        size = self.read_uint32()
        if self._last_error:
            return string()
        if size + self._pos > self._buf_len:
            self._pos = old_pos
            self._last_error = BD_BUFFER_EXHAUSTED
            return string()

        return string(self._buf_ptr + self._pos, size)

    cdef int _read_input_byte(self) nogil:
        cdef int ret
        if self._pos < self._buf_len:
            self._last_error = BD_SUCCESS
            ret = self._buf_ptr[self._pos]
            self._pos += 1
            return ret
        else:
            self._last_error = BD_BUFFER_EXHAUSTED
            return 0

    cdef int32_t _get_varint32(self) nogil:
        """
        Deserialize a protobuf varint read from input stream; update
        offset based on number of bytes consumed.
        """
        cdef int32_t value = 0
        cdef int32_t base = 1
        cdef int index = 0
        cdef int val_byte

        while True:
            val_byte = self._read_input_byte()
            if self._last_error != BD_SUCCESS:
                return 0
            value += (val_byte & 0x7F) * base
            if val_byte & 0x80:
                base *= 128
            else:
                return value

    cdef int64_t _get_varint64(self) nogil:
        """
        Deserialize a protobuf varint read from input stream; update
        offset based on number of bytes consumed.
        """
        cdef int64_t value = 0
        cdef int64_t base = 1
        cdef int index = 0
        cdef int val_byte

        while True:
            val_byte = self._read_input_byte()
            if self._last_error != BD_SUCCESS:
                return 0
            value += (val_byte & 0x7F) * base
            if val_byte & 0x80:
                base *= 128
            else:
                return value

    cdef int32_t _get_signed_varint32(self) nogil:
        """
        Deserialize a signed protobuf varint read from input stream;
        update offset based on number of bytes consumed.
        """
        cdef uint32_t value = 0
        cdef int32_t base = 1
        cdef int index = 0
        cdef int val_byte

        while True:
            val_byte = self._read_input_byte()
            if self._last_error != BD_SUCCESS:
                return 0
            value += (val_byte & 0x7F) * base
            if val_byte & 0x80:
                base *= 128
            else:
                return <int32_t>((value >> 1) ^ (-(value & 1))) # zigzag decoding

    cdef int64_t _get_signed_varint64(self) nogil:
        """
        Deserialize a signed protobuf varint read from input stream;
        update offset based on number of bytes consumed.
        """
        cdef uint64_t value = 0
        cdef int64_t base = 1
        cdef int index = 0
        cdef int val_byte

        while True:
            val_byte = self._read_input_byte()
            if self._last_error != BD_SUCCESS:
                return 0
            value += (val_byte & 0x7F) * base
            if val_byte & 0x80:
                base *= 128
            else:
                return <int64_t>((value >> 1) ^ (-(value & 1))) # zigzag decoding



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2022 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ...compat import Enum, six
from ...errors import DependencyNotInstalledError
from ..expr.expressions import CollectionExpr, SequenceExpr, Expr
from ..expr.dynamic import DynamicMixin
from ..types import is_number


class PlottingCore(Enum):
    PANDAS = 'pandas'


def _plot_pandas(df, method='plot', **kwargs):
    x_label, y_label = kwargs.pop('xlabel', None), kwargs.pop('ylabel', None)

    x_label_size, y_label_size = kwargs.pop('xlabelsize', None), kwargs.pop('ylabelsize', None)
    label_size = kwargs.pop('labelsize', None)
    x_label_size = x_label_size or label_size
    y_label_size = y_label_size or label_size

    title_size = kwargs.pop('titlesize', None)
    title = kwargs.pop('title', None)

    annotate = kwargs.pop('annotate', None)
    x_annotate_scale = kwargs.pop('xannotatescale', 1.0)
    y_annotate_scale = kwargs.pop('yannotatescale', 1.0)

    fig = getattr(df, method)(**kwargs)

    import numpy as np
    if isinstance(fig, np.ndarray):
        figs = fig
        fig = fig[0]
    else:
        figs = [fig, ]

    if x_label:
        if x_label_size:
            fig.set_xlabel(x_label, fontsize=x_label_size)
        else:
            fig.set_xlabel(x_label)
    if y_label:
        if y_label_size:
            fig.set_ylabel(y_label, fontsize=y_label_size)
        else:
            fig.set_ylabel(y_label)

    if title_size:
        fig.title.set_fontsize(title_size)
    if title:
        fig.set_title(title)

    if annotate:
        for ax in figs:
            for p in ax.patches:
                ax.annotate(str(p.get_height()), (p.get_x(), p.get_height()),
                            xytext=(p.get_x() * (x_annotate_scale - 1),
                                    p.get_height() * (y_annotate_scale - 1)),
                            textcoords='offset points')
            for l in ax.lines:
                xs, ys = l.get_data()
                for x, y in zip(xs, ys):
                    ax.annotate(str(y), (x, y),
                                xytext=(x * (x_annotate_scale - 1), y * (y_annotate_scale - 1)),
                                textcoords='offset points')

    return fig


def _hist_pandas(df, **kwargs):
    return _plot_pandas(df, method='hist', **kwargs)


def _boxplot_pandas(df, **kwargs):
    return _plot_pandas(df, method='boxplot', **kwargs)


def _plot_sequence(expr, kind='line', use_cache=None, **kwargs):
    try:
        import pandas as pd
    except ImportError:
        raise DependencyNotInstalledError('plot requires for pandas')

    series = expr.to_pandas(use_cache=use_cache)
    for k in list(kwargs.keys()):
        v = kwargs[k]
        if v is not None and isinstance(v, (CollectionExpr, SequenceExpr)):
            kwargs[k] = v.to_pandas()
    return _plot_pandas(series, kind=kind, **kwargs)


def _hist_sequence(expr, use_cache=None, **kwargs):
    try:
        import pandas as pd
    except ImportError:
        raise DependencyNotInstalledError('plot requires for pandas')

    series = expr.to_pandas(use_cache=use_cache)
    return _hist_pandas(series, **kwargs)


def _plot_collection(expr, x=None, y=None, kind='line', **kwargs):
    try:
        import pandas as pd
    except ImportError:
        raise DependencyNotInstalledError('plot requires for pandas')

    if isinstance(expr, DynamicMixin):
        rf = expr.head(1)
        columns = rf.columns
    else:
        columns = expr.dtypes

    fields = []
    x_name = None
    y_name = None

    if x is not None:
        fields.append(x)
        if isinstance(x, six.string_types):
            x_name = x
        else:
            x_name = x.name
    if y is not None:
        fields.append(y)
        if isinstance(y, six.string_types):
            y_name = y
        else:
            y_name = y.name

    if x_name is None or y_name is None:
        for col in columns:
            if col.name == x_name or col.name == y_name:
                continue
            elif is_number(col.type):
                fields.append(col.name)

    to_replace = dict()
    for k, v in six.iteritems(kwargs):
        if isinstance(v, Expr):
            fields.append(v)
            to_replace[k] = v.name

    df = expr[fields].to_pandas()

    if x_name is not None:
        kwargs['x'] = x_name
    if y is not None:
        kwargs['y'] = y_name

    for k, v in six.iteritems(to_replace):
        kwargs[k] = df[v]

    _plot_func = kwargs.pop('plot_func', _plot_pandas)
    return _plot_func(df, kind=kind, **kwargs)


def _hist_collection(expr, **kwargs):
    try:
        import pandas as pd
    except ImportError:
        raise DependencyNotInstalledError('plot requires for pandas')

    fields = []
    column = kwargs.get('column')
    if isinstance(column, six.string_types):
        fields.append(column)

    to_replace = dict()
    for k, v in six.iteritems(kwargs):
        if isinstance(v, Expr):
            fields.append(v)
            to_replace[k] = v.name

    if fields:
        expr = expr[fields]

    df = expr.to_pandas()

    for k, v in six.iteritems(to_replace):
        kwargs[k] = df[v]

    _plot_func = kwargs.pop('plot_func', _hist_pandas)
    return _plot_func(df, **kwargs)


def _boxplot_collection(expr, **kwargs):
    try:
        import pandas as pd
    except ImportError:
        raise DependencyNotInstalledError('plot requires for pandas')

    fields = set()

    column = kwargs.get('column')
    if isinstance(column, six.string_types):
        fields.add(column)
    elif column is not None:
        fields = fields.union([column])

    by = kwargs.get('by')
    if isinstance(by, six.string_types):
        fields.add(by)
    elif by is not None:
        fields = fields.union([by])

    fields = list(fields)
    to_replace = dict()
    for k, v in six.iteritems(kwargs):
        if isinstance(v, Expr):
            fields.append(v)
            to_replace[k] = v.name

    if fields:
        expr = expr[fields]

    df = expr.to_pandas()

    for k, v in six.iteritems(to_replace):
        kwargs[k] = df[v]

    _plot_func = kwargs.pop('plot_func', _boxplot_pandas)
    return _plot_func(df, **kwargs)


CollectionExpr.plot = _plot_collection
CollectionExpr.hist = _hist_collection
CollectionExpr.boxplot = _boxplot_collection
SequenceExpr.plot = _plot_sequence
SequenceExpr.hist = _hist_sequence

try:
    from pandas.tools.plotting import plot_frame, hist_frame, boxplot, \
        plot_series, hist_series

    _plot_collection.__doc__ = plot_frame.__doc__
    _hist_collection.__doc__ = hist_frame.__doc__
    _boxplot_collection.__doc__ = boxplot.__doc__
    _plot_sequence.__doc__ = plot_series.__doc__
    _hist_sequence.__doc__ = hist_series.__doc__
except ImportError:
    pass

__all__ = ()



# encoding: utf-8
# Copyright 1999-2022 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import

import functools
import json
import re
import time
import uuid
from collections import namedtuple, OrderedDict

from odps.df.expr.dynamic import DynamicMixin
from .objects import SchemaDef
from ..expr import ModelDataCollectionExpr, ODPSModelExpr
from ..expr.op import ProgrammaticFieldChangeOperation
from ..pipeline.core import PipelineStep
from ..utils import import_class_member, get_function_args, ML_ARG_PREFIX
from ...compat import six, irange, Iterable
from ...df.core import DataFrame
from ...models import TableSchema
from ..enums import PortType, PortDirection
from ...utils import underline_to_capitalized, camel_to_underline, survey


def is_df_object(obj):
    from ...df.core import CollectionExpr
    from ...df.expr.expressions import SequenceExpr
    return isinstance(obj, (CollectionExpr, SequenceExpr))


def is_ml_object(obj):
    return is_df_object(obj) or isinstance(obj, ODPSModelExpr)


class BaseAlgorithm(PipelineStep):
    _entry_method = None

    def __init__(self, name, parameters, ports, metas=None):
        self._name = name
        self._parameters = parameters
        self._ports = ports
        self._metas = metas if metas is not None else {}
        if not hasattr(self, '_reload_ml_fields'):
            self._reload_ml_fields = False

        output_names = [p.name for p in sorted(six.itervalues(self._ports), key=lambda pt: pt.sequence)
                        if p.io_type == PortDirection.OUTPUT]
        param_names = [p for p in six.iterkeys(self._parameters)]
        [self._invoke_setter(p.name, p.value) for p in six.itervalues(self._parameters)]

        super(BaseAlgorithm, self).__init__(name, param_names=param_names, output_names=output_names)

    @staticmethod
    def _get_exec_id():
        return '%d-%s' % (int(time.time()), str(uuid.uuid4()))

    def _invoke_setter(self, param_name, value):
        if not self._parameters[param_name].setter:
            return
        method = import_class_member(self._parameters[param_name].setter)
        method(self, param_name, value)

    def _map_inputs_from_args(self, *args, **kwargs):
        inputs = [p for p in sorted(six.itervalues(self._ports), key=lambda pt: pt.sequence)
                  if p.io_type == PortDirection.INPUT]

        args = [arg for arg in args if is_ml_object(arg)]
        ml_kw = dict((k, v) for k, v in six.iteritems(kwargs) if is_ml_object(v))

        total_input = len(args) + len(ml_kw)
        if total_input == 0:
            raise TypeError('Can only perform transformation on DataFrames')
        if total_input > len(inputs):
            raise ValueError('Input count mismatch.')

        # combine args
        obj_dict = dict(zip((p.name for p in inputs), args))
        obj_dict.update(ml_kw)

        return obj_dict

    def _build_collection(self, input_obj_dict, schema_def, coll_type=None, **expr_kw):
        coll_type = coll_type or self._collection_expr
        if schema_def and schema_def.dynamic:
            coll_type = type(coll_type)('Dynamic' + coll_type.__name__, (DynamicMixin, coll_type), {})

        if schema_def is None:
            schema_def = SchemaDef()
            schema_def.copy_input = six.next(k for k, v in six.iteritems(input_obj_dict) if is_df_object(v))

        kw = dict((ML_ARG_PREFIX + k, v) for k, v in six.iteritems(input_obj_dict))
        kw.update(expr_kw)
        if schema_def.copy_input is not None:
            src_df = coll_type(register_expr=True, _schema=input_obj_dict[schema_def.copy_input]._schema, **kw)
            src_df._ml_uplink = [input_obj_dict[schema_def.copy_input]]
        else:
            src_df = coll_type(register_expr=True, _schema=TableSchema(_columns=[]), **kw)
            src_df._ml_uplink = [o for o in six.itervalues(input_obj_dict) if is_df_object(o)]

        out_schemas = dict((pname, dict((f.name, f.type) for f in df._ml_fields))
                           for pname, df in six.iteritems(input_obj_dict) if is_df_object(df))
        for model_name, model in ((nm, m) for nm, m in six.iteritems(input_obj_dict) if isinstance(m, ODPSModelExpr)):
            if not getattr(model, '_model_collections', None):
                continue
            out_schemas.update(dict(('%s.%s' % (model_name, ds_name), dict((f.name, f.type) for f in df._ml_fields))
                                    for ds_name, df in six.iteritems(model._model_collections)))

        if schema_def.programmatic:
            generator = import_class_member(schema_def.schema)
        else:
            generator = _static_ml_fields_generator
        func_args = get_function_args(generator)

        kw = OrderedDict()
        if 'params' in func_args:
            kw['params'] = src_df.convert_params()
        if 'fields' in func_args:
            kw['fields'] = out_schemas
        if 'algorithm' in func_args:
            kw['algorithm'] = self
        if 'schema' in func_args:
            kw['schema'] = schema_def.schema

        op = ProgrammaticFieldChangeOperation(functools.partial(generator, **kw), schema_def.copy_input is not None)
        src_df._perform_operation(op)
        src_df._rebuild_df_schema(schema_def.dynamic)
        return src_df

    @survey
    def _do_transform(self, *args, **kwargs):
        exec_id = self._get_exec_id()

        out_port_defs = [p for p in sorted(six.itervalues(self._ports), key=lambda pt: pt.sequence)
                         if p.io_type == PortDirection.OUTPUT]
        params = dict((k, v.value) for k, v in six.iteritems(self._parameters))

        obj_dict = self._map_inputs_from_args(*args, **kwargs)
        # pick out engine kwargs such as core number, mem usage, etc.
        engine_kw = dict(p for p in six.iteritems(kwargs) if p[0] not in obj_dict)

        out_objs = dict()
        for port_def in out_port_defs:
            port_param_name = camel_to_underline(port_def.name)

            expr_kw = dict(_params=params, _exec_id=exec_id, _output_name=port_def.name,
                           _engine_kw=engine_kw)
            if port_def.type == PortType.DATA:
                out_objs[port_param_name] = self._build_collection(obj_dict, port_def.schema, **expr_kw)
            elif port_def.type == PortType.MODEL:
                expr_kw.update((ML_ARG_PREFIX + k, v) for k, v in six.iteritems(obj_dict))

                model_type = port_def.model.type if port_def.model else 'PmmlModel'

                if model_type == 'PmmlModel':
                    expr_kw['_is_offline_model'] = True
                else:
                    expr_kw['_is_offline_model'] = False
                    expr_kw['_model_collections'] = None

                out_objs[port_param_name] = model_expr = self._model_expr(register_expr=True, **expr_kw)

                if model_type != 'PmmlModel':
                    dfs = dict()
                    for schema in port_def.model.schemas:
                        if schema.direct_copy:
                            dfs[schema.name] = obj_dict[schema.direct_copy].view()
                        else:
                            ds = self._build_collection(obj_dict, schema, coll_type=ModelDataCollectionExpr,
                                                        _mlattr_model=out_objs[port_param_name], _data_item=schema.name,
                                                        model_inputs=list(six.iterkeys(obj_dict)))
                            dfs[schema.name] = ds
                    model_expr._model_collections = dfs

                    if port_def.model and hasattr(port_def.model, 'copy_params') and port_def.model.copy_params:
                        param_dict = dict((self._parameters[pn].name, self._parameters[pn].value)
                                          for pn in port_def.model.copy_params.split(','))
                        model_expr._model_params.update(param_dict)

            if port_def.exporter:
                out_obj = out_objs[port_param_name]
                method = import_class_member(port_def.exporter)
                out_objs[port_param_name] = method(out_obj)

        result_type = namedtuple(underline_to_capitalized(self._name) + 'Result',
                                 [camel_to_underline(p.name) for p in out_port_defs])
        ret_tuple = result_type(**out_objs)
        if len(ret_tuple) == 1:
            return ret_tuple[0]
        else:
            return ret_tuple


class BaseProcessAlgorithm(BaseAlgorithm):
    _entry_method = 'transform'
    """
    Base processing algorithm class
    """
    def transform(self, *args, **kwargs):
        """
        :type args: list[DataFrame]
        """
        return self._do_transform(*args, **kwargs)


class BaseTrainingAlgorithm(BaseAlgorithm):
    _entry_method = 'train'
    """
    Base class for supervised clustering algorithms
    """
    def train(self, *args, **kwargs):
        """
        Perform training on a DataFrame.
        The label field is specified by the ``label_field`` method.

        :param train_data: DataFrame to be trained. Label field must be specified.
        :type train_data: DataFrame

        :return: Trained model
        :rtype: MLModel
        """
        objs = self._do_transform(*args, **kwargs)
        obj_list = [objs, ] if not isinstance(objs, Iterable) else objs
        for obj in obj_list:
            if not isinstance(obj, ODPSModelExpr):
                continue
            for meta in ['predictor', 'recommender']:
                if meta not in self._metas:
                    continue
                mod = __import__(self.__class__.__module__.__name__, fromlist=[''])\
                    if not hasattr(self, '_env') else self._env
                action_cls_name = underline_to_capitalized(self._metas[meta])
                if not hasattr(mod, action_cls_name):
                    action_cls_name = '_' + action_cls_name
                setattr(obj, '_' + meta, mod + '.' + action_cls_name)

        return objs


class BaseMetricsAlgorithm(BaseAlgorithm):
    def calc(self, *args, **kwargs):
        """
        :type args: list[DataFrame]
        """
        cases = kwargs.pop('_cases', [])
        if not isinstance(cases, Iterable):
            cases = [cases, ]
        result_callback = kwargs.pop('_result_callback', None)
        execute_now = kwargs.pop('execute_now', True)

        exec_id = self._get_exec_id()
        params = dict((k, v.value) for k, v in six.iteritems(self._parameters))

        obj_dict = self._map_inputs_from_args(*args, **kwargs)
        engine_kw = dict(p for p in six.iteritems(kwargs) if p[0] not in obj_dict)

        expr_kw = dict(_params=params, _exec_id=exec_id, _engine_kw=engine_kw)
        if cases:
            expr_kw['_cases'] = cases

        expr_kw.update((ML_ARG_PREFIX + k, v) for k, v in six.iteritems(obj_dict))
        if result_callback:
            expr_kw['_result_callback'] = result_callback
        expr = self._metrics_expr(**expr_kw)

        if execute_now:
            return expr.execute()
        else:
            return expr

_FIELD_PARAM_PLACEHOLDER = re.compile(r'\{([^\}]+)\}')
_REDUNDANT_COMMA = re.compile(r',\s*,')


def _static_ml_fields_generator(params, fields, algorithm, schema):

    def format_col(field_name, field_type, **kw):
        prefix = kw.get('prefix', '')
        role = kw.get('role')
        if role:
            return '%s:%s:%s' % (prefix + field_name, field_type, role)
        else:
            return '%s:%s' % (prefix + field_name, field_type)

    if isinstance(schema, list):
        name_list = []
        for f in schema:
            fr = '{0}:{1}'.format(f.name, f.type)
            if f.role:
                fr += ':' + '#'.join(r.name for r in f.role)
            name_list.append(fr)
        return ','.join(name_list)
    elif schema is None:
        return ''
    else:
        def _sub_func(m):
            cmd = [s.strip() for s in m.group(1).split('|')]
            cmd_name = cmd[0]
            if cmd_name.startswith('#'):
                cmd_name = cmd_name[1:]
                if cmd_name == 'range':
                    for idx, c in enumerate(cmd[1:]):
                        for k, v in six.iteritems(params):
                            c = c.replace(k, str(v))
                        cmd[idx + 1] = c
                    start_expr, end_expr, tmpl = cmd[1:]
                    start_num = json.loads(start_expr)
                    end_num = json.loads(end_expr)
                    flist = [tmpl.replace('#n', str(n)) for n in irange(start_num, end_num)]
                    return ','.join(flist)
                elif cmd_name == 'input':
                    format_args = dict(tuple(p.split('=', 1)) for p in cmd[2:])
                    input_ml_fields = fields[cmd[1]]
                    return ','.join(format_col(col, typ, **format_args) for col, typ in six.iteritems(input_ml_fields))
                else:
                    raise ValueError('Unsupported method')
            else:
                format_args = dict(tuple(p.split('=', 1)) for p in cmd[1:])

                input_name = algorithm._parameters[cmd_name].input_name
                if not input_name:
                    return params[cmd_name]
                cols = params[cmd_name] if cmd_name in params else []
                if isinstance(cols, six.string_types):
                    cols = cols.split(',')
                input_ml_fields = fields[input_name]
                cols = cols or []

                return ','.join(format_col(col, input_ml_fields[col], **format_args) for col in cols)
        return _REDUNDANT_COMMA.sub(',', _FIELD_PARAM_PLACEHOLDER.sub(_sub_func, schema))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2022 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


class ExecuteContext(object):
    def __init__(self):
        self._expr_id_cached_data = dict()

    def cache(self, expr, data):
        if data is not None:
            self._expr_id_cached_data[expr._id] = data

    def is_cached(self, expr):
        return expr._id in self._expr_id_cached_data

    def get_cached(self, expr):
        return self._expr_id_cached_data[expr._id]

    def uncache(self, expr):
        if expr._id in self._expr_id_cached_data:
            del self._expr_id_cached_data[expr._id]

context = ExecuteContext()



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: cupid_subprocess_service.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
from google.protobuf import service as _service
from google.protobuf import service_reflection
from google.protobuf import descriptor_pb2
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='cupid_subprocess_service.proto',
  package='apsara.odps.cupid.protocol',
  serialized_pb=_b('\n\x1e\x63upid_subprocess_service.proto\x12\x1a\x61psara.odps.cupid.protocol\"\xb7\x01\n\nInputSplit\x12\x14\n\x0csplitIndexId\x18\x01 \x01(\r\x12\x16\n\x0esplitFileStart\x18\x02 \x01(\x04\x12\x14\n\x0csplitFileEnd\x18\x03 \x01(\x04\x12\x17\n\x0fschemaFileStart\x18\x04 \x01(\x04\x12\x15\n\rschemaFileEnd\x18\x05 \x01(\x04\x12\x0f\n\x07project\x18\x06 \x01(\t\x12\r\n\x05table\x18\x07 \x01(\t\x12\x15\n\rpartitionSpec\x18\x08 \x01(\t\"3\n\x0eInputSplitMeta\x12\x10\n\x08rowCount\x18\x01 \x01(\x04\x12\x0f\n\x07rawSize\x18\x02 \x01(\x04\"0\n\x14GetSplitsMetaRequest\x12\x18\n\x10inputTableHandle\x18\x01 \x01(\t\"\\\n\x15GetSplitsMetaResponse\x12\x43\n\x0finputSplitsMeta\x18\x01 \x03(\x0b\x32*.apsara.odps.cupid.protocol.InputSplitMeta\",\n\x10GetSplitsRequest\x12\x18\n\x10inputTableHandle\x18\x01 \x01(\t\"P\n\x11GetSplitsResponse\x12;\n\x0binputSplits\x18\x01 \x03(\x0b\x32&.apsara.odps.cupid.protocol.InputSplit\"r\n\x1aRegisterTableReaderRequest\x12\x18\n\x10inputTableHandle\x18\x01 \x01(\t\x12:\n\ninputSplit\x18\x02 \x01(\x0b\x32&.apsara.odps.cupid.protocol.InputSplit\"\\\n\x1bRegisterTableReaderResponse\x12\x0e\n\x06schema\x18\x01 \x01(\t\x12\x14\n\x0creadIterator\x18\x02 \x01(\t\x12\x17\n\x0fpartitionSchema\x18\x03 \x01(\t\"\xb5\x01\n\x1aRegisterTableWriterRequest\x12\x19\n\x11outputTableHandle\x18\x01 \x01(\t\x12\x13\n\x0bprojectName\x18\x02 \x01(\t\x12\x11\n\ttableName\x18\x03 \x01(\t\x12\x17\n\x0f\x61ttemptFileName\x18\x04 \x01(\t\x12\x10\n\x08partSpec\x18\x05 \x01(\t\x12\x0e\n\x06schema\x18\x06 \x01(\t\x12\x19\n\x11tableMetaFileName\x18\x07 \x01(\t\"@\n\x1bRegisterTableWriterResponse\x12!\n\x19subprocessWriteTableLabel\x18\x01 \x01(\t\"U\n\x10\x43ommitActionInfo\x12\x10\n\x08partSpec\x18\x01 \x01(\t\x12\x17\n\x0f\x61ttemptFileName\x18\x02 \x01(\t\x12\x16\n\x0e\x63ommitFileName\x18\x03 \x01(\t\"\xa5\x01\n\x17\x43ommitTableFilesRequest\x12\x19\n\x11outputTableHandle\x18\x01 \x01(\t\x12\x13\n\x0bprojectName\x18\x02 \x01(\t\x12\x11\n\ttableName\x18\x03 \x01(\t\x12G\n\x11\x63ommitActionInfos\x18\x04 \x03(\x0b\x32,.apsara.odps.cupid.protocol.CommitActionInfo\"\x1a\n\x18\x43ommitTableFilesResponse2\x89\x05\n\x16\x43upidSubProcessService\x12h\n\tGetSplits\x12,.apsara.odps.cupid.protocol.GetSplitsRequest\x1a-.apsara.odps.cupid.protocol.GetSplitsResponse\x12\x86\x01\n\x13RegisterTableReader\x12\x36.apsara.odps.cupid.protocol.RegisterTableReaderRequest\x1a\x37.apsara.odps.cupid.protocol.RegisterTableReaderResponse\x12\x86\x01\n\x13RegisterTableWriter\x12\x36.apsara.odps.cupid.protocol.RegisterTableWriterRequest\x1a\x37.apsara.odps.cupid.protocol.RegisterTableWriterResponse\x12}\n\x10\x43ommitTableFiles\x12\x33.apsara.odps.cupid.protocol.CommitTableFilesRequest\x1a\x34.apsara.odps.cupid.protocol.CommitTableFilesResponse\x12t\n\rGetSplitsMeta\x12\x30.apsara.odps.cupid.protocol.GetSplitsMetaRequest\x1a\x31.apsara.odps.cupid.protocol.GetSplitsMetaResponseB&B\x1b\x43upidSubProcessServiceProto\x80\x01\x01\x88\x01\x01\x90\x01\x01')
)
_sym_db.RegisterFileDescriptor(DESCRIPTOR)




_INPUTSPLIT = _descriptor.Descriptor(
  name='InputSplit',
  full_name='apsara.odps.cupid.protocol.InputSplit',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='splitIndexId', full_name='apsara.odps.cupid.protocol.InputSplit.splitIndexId', index=0,
      number=1, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='splitFileStart', full_name='apsara.odps.cupid.protocol.InputSplit.splitFileStart', index=1,
      number=2, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='splitFileEnd', full_name='apsara.odps.cupid.protocol.InputSplit.splitFileEnd', index=2,
      number=3, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='schemaFileStart', full_name='apsara.odps.cupid.protocol.InputSplit.schemaFileStart', index=3,
      number=4, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='schemaFileEnd', full_name='apsara.odps.cupid.protocol.InputSplit.schemaFileEnd', index=4,
      number=5, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='project', full_name='apsara.odps.cupid.protocol.InputSplit.project', index=5,
      number=6, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='table', full_name='apsara.odps.cupid.protocol.InputSplit.table', index=6,
      number=7, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='partitionSpec', full_name='apsara.odps.cupid.protocol.InputSplit.partitionSpec', index=7,
      number=8, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=63,
  serialized_end=246,
)


_INPUTSPLITMETA = _descriptor.Descriptor(
  name='InputSplitMeta',
  full_name='apsara.odps.cupid.protocol.InputSplitMeta',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='rowCount', full_name='apsara.odps.cupid.protocol.InputSplitMeta.rowCount', index=0,
      number=1, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='rawSize', full_name='apsara.odps.cupid.protocol.InputSplitMeta.rawSize', index=1,
      number=2, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=248,
  serialized_end=299,
)


_GETSPLITSMETAREQUEST = _descriptor.Descriptor(
  name='GetSplitsMetaRequest',
  full_name='apsara.odps.cupid.protocol.GetSplitsMetaRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='inputTableHandle', full_name='apsara.odps.cupid.protocol.GetSplitsMetaRequest.inputTableHandle', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=301,
  serialized_end=349,
)


_GETSPLITSMETARESPONSE = _descriptor.Descriptor(
  name='GetSplitsMetaResponse',
  full_name='apsara.odps.cupid.protocol.GetSplitsMetaResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='inputSplitsMeta', full_name='apsara.odps.cupid.protocol.GetSplitsMetaResponse.inputSplitsMeta', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=351,
  serialized_end=443,
)


_GETSPLITSREQUEST = _descriptor.Descriptor(
  name='GetSplitsRequest',
  full_name='apsara.odps.cupid.protocol.GetSplitsRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='inputTableHandle', full_name='apsara.odps.cupid.protocol.GetSplitsRequest.inputTableHandle', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=445,
  serialized_end=489,
)


_GETSPLITSRESPONSE = _descriptor.Descriptor(
  name='GetSplitsResponse',
  full_name='apsara.odps.cupid.protocol.GetSplitsResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='inputSplits', full_name='apsara.odps.cupid.protocol.GetSplitsResponse.inputSplits', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=491,
  serialized_end=571,
)


_REGISTERTABLEREADERREQUEST = _descriptor.Descriptor(
  name='RegisterTableReaderRequest',
  full_name='apsara.odps.cupid.protocol.RegisterTableReaderRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='inputTableHandle', full_name='apsara.odps.cupid.protocol.RegisterTableReaderRequest.inputTableHandle', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='inputSplit', full_name='apsara.odps.cupid.protocol.RegisterTableReaderRequest.inputSplit', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=573,
  serialized_end=687,
)


_REGISTERTABLEREADERRESPONSE = _descriptor.Descriptor(
  name='RegisterTableReaderResponse',
  full_name='apsara.odps.cupid.protocol.RegisterTableReaderResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='schema', full_name='apsara.odps.cupid.protocol.RegisterTableReaderResponse.schema', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='readIterator', full_name='apsara.odps.cupid.protocol.RegisterTableReaderResponse.readIterator', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='partitionSchema', full_name='apsara.odps.cupid.protocol.RegisterTableReaderResponse.partitionSchema', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=689,
  serialized_end=781,
)


_REGISTERTABLEWRITERREQUEST = _descriptor.Descriptor(
  name='RegisterTableWriterRequest',
  full_name='apsara.odps.cupid.protocol.RegisterTableWriterRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='outputTableHandle', full_name='apsara.odps.cupid.protocol.RegisterTableWriterRequest.outputTableHandle', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='projectName', full_name='apsara.odps.cupid.protocol.RegisterTableWriterRequest.projectName', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='tableName', full_name='apsara.odps.cupid.protocol.RegisterTableWriterRequest.tableName', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='attemptFileName', full_name='apsara.odps.cupid.protocol.RegisterTableWriterRequest.attemptFileName', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='partSpec', full_name='apsara.odps.cupid.protocol.RegisterTableWriterRequest.partSpec', index=4,
      number=5, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='schema', full_name='apsara.odps.cupid.protocol.RegisterTableWriterRequest.schema', index=5,
      number=6, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='tableMetaFileName', full_name='apsara.odps.cupid.protocol.RegisterTableWriterRequest.tableMetaFileName', index=6,
      number=7, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=784,
  serialized_end=965,
)


_REGISTERTABLEWRITERRESPONSE = _descriptor.Descriptor(
  name='RegisterTableWriterResponse',
  full_name='apsara.odps.cupid.protocol.RegisterTableWriterResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='subprocessWriteTableLabel', full_name='apsara.odps.cupid.protocol.RegisterTableWriterResponse.subprocessWriteTableLabel', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=967,
  serialized_end=1031,
)


_COMMITACTIONINFO = _descriptor.Descriptor(
  name='CommitActionInfo',
  full_name='apsara.odps.cupid.protocol.CommitActionInfo',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='partSpec', full_name='apsara.odps.cupid.protocol.CommitActionInfo.partSpec', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='attemptFileName', full_name='apsara.odps.cupid.protocol.CommitActionInfo.attemptFileName', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='commitFileName', full_name='apsara.odps.cupid.protocol.CommitActionInfo.commitFileName', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1033,
  serialized_end=1118,
)


_COMMITTABLEFILESREQUEST = _descriptor.Descriptor(
  name='CommitTableFilesRequest',
  full_name='apsara.odps.cupid.protocol.CommitTableFilesRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='outputTableHandle', full_name='apsara.odps.cupid.protocol.CommitTableFilesRequest.outputTableHandle', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='projectName', full_name='apsara.odps.cupid.protocol.CommitTableFilesRequest.projectName', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='tableName', full_name='apsara.odps.cupid.protocol.CommitTableFilesRequest.tableName', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='commitActionInfos', full_name='apsara.odps.cupid.protocol.CommitTableFilesRequest.commitActionInfos', index=3,
      number=4, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1121,
  serialized_end=1286,
)


_COMMITTABLEFILESRESPONSE = _descriptor.Descriptor(
  name='CommitTableFilesResponse',
  full_name='apsara.odps.cupid.protocol.CommitTableFilesResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1288,
  serialized_end=1314,
)

_GETSPLITSMETARESPONSE.fields_by_name['inputSplitsMeta'].message_type = _INPUTSPLITMETA
_GETSPLITSRESPONSE.fields_by_name['inputSplits'].message_type = _INPUTSPLIT
_REGISTERTABLEREADERREQUEST.fields_by_name['inputSplit'].message_type = _INPUTSPLIT
_COMMITTABLEFILESREQUEST.fields_by_name['commitActionInfos'].message_type = _COMMITACTIONINFO
DESCRIPTOR.message_types_by_name['InputSplit'] = _INPUTSPLIT
DESCRIPTOR.message_types_by_name['InputSplitMeta'] = _INPUTSPLITMETA
DESCRIPTOR.message_types_by_name['GetSplitsMetaRequest'] = _GETSPLITSMETAREQUEST
DESCRIPTOR.message_types_by_name['GetSplitsMetaResponse'] = _GETSPLITSMETARESPONSE
DESCRIPTOR.message_types_by_name['GetSplitsRequest'] = _GETSPLITSREQUEST
DESCRIPTOR.message_types_by_name['GetSplitsResponse'] = _GETSPLITSRESPONSE
DESCRIPTOR.message_types_by_name['RegisterTableReaderRequest'] = _REGISTERTABLEREADERREQUEST
DESCRIPTOR.message_types_by_name['RegisterTableReaderResponse'] = _REGISTERTABLEREADERRESPONSE
DESCRIPTOR.message_types_by_name['RegisterTableWriterRequest'] = _REGISTERTABLEWRITERREQUEST
DESCRIPTOR.message_types_by_name['RegisterTableWriterResponse'] = _REGISTERTABLEWRITERRESPONSE
DESCRIPTOR.message_types_by_name['CommitActionInfo'] = _COMMITACTIONINFO
DESCRIPTOR.message_types_by_name['CommitTableFilesRequest'] = _COMMITTABLEFILESREQUEST
DESCRIPTOR.message_types_by_name['CommitTableFilesResponse'] = _COMMITTABLEFILESRESPONSE

InputSplit = _reflection.GeneratedProtocolMessageType('InputSplit', (_message.Message,), dict(
  DESCRIPTOR = _INPUTSPLIT,
  __module__ = 'cupid_subprocess_service_pb2'
  # @@protoc_insertion_point(class_scope:apsara.odps.cupid.protocol.InputSplit)
  ))
_sym_db.RegisterMessage(InputSplit)

InputSplitMeta = _reflection.GeneratedProtocolMessageType('InputSplitMeta', (_message.Message,), dict(
  DESCRIPTOR = _INPUTSPLITMETA,
  __module__ = 'cupid_subprocess_service_pb2'
  # @@protoc_insertion_point(class_scope:apsara.odps.cupid.protocol.InputSplitMeta)
  ))
_sym_db.RegisterMessage(InputSplitMeta)

GetSplitsMetaRequest = _reflection.GeneratedProtocolMessageType('GetSplitsMetaRequest', (_message.Message,), dict(
  DESCRIPTOR = _GETSPLITSMETAREQUEST,
  __module__ = 'cupid_subprocess_service_pb2'
  # @@protoc_insertion_point(class_scope:apsara.odps.cupid.protocol.GetSplitsMetaRequest)
  ))
_sym_db.RegisterMessage(GetSplitsMetaRequest)

GetSplitsMetaResponse = _reflection.GeneratedProtocolMessageType('GetSplitsMetaResponse', (_message.Message,), dict(
  DESCRIPTOR = _GETSPLITSMETARESPONSE,
  __module__ = 'cupid_subprocess_service_pb2'
  # @@protoc_insertion_point(class_scope:apsara.odps.cupid.protocol.GetSplitsMetaResponse)
  ))
_sym_db.RegisterMessage(GetSplitsMetaResponse)

GetSplitsRequest = _reflection.GeneratedProtocolMessageType('GetSplitsRequest', (_message.Message,), dict(
  DESCRIPTOR = _GETSPLITSREQUEST,
  __module__ = 'cupid_subprocess_service_pb2'
  # @@protoc_insertion_point(class_scope:apsara.odps.cupid.protocol.GetSplitsRequest)
  ))
_sym_db.RegisterMessage(GetSplitsRequest)

GetSplitsResponse = _reflection.GeneratedProtocolMessageType('GetSplitsResponse', (_message.Message,), dict(
  DESCRIPTOR = _GETSPLITSRESPONSE,
  __module__ = 'cupid_subprocess_service_pb2'
  # @@protoc_insertion_point(class_scope:apsara.odps.cupid.protocol.GetSplitsResponse)
  ))
_sym_db.RegisterMessage(GetSplitsResponse)

RegisterTableReaderRequest = _reflection.GeneratedProtocolMessageType('RegisterTableReaderRequest', (_message.Message,), dict(
  DESCRIPTOR = _REGISTERTABLEREADERREQUEST,
  __module__ = 'cupid_subprocess_service_pb2'
  # @@protoc_insertion_point(class_scope:apsara.odps.cupid.protocol.RegisterTableReaderRequest)
  ))
_sym_db.RegisterMessage(RegisterTableReaderRequest)

RegisterTableReaderResponse = _reflection.GeneratedProtocolMessageType('RegisterTableReaderResponse', (_message.Message,), dict(
  DESCRIPTOR = _REGISTERTABLEREADERRESPONSE,
  __module__ = 'cupid_subprocess_service_pb2'
  # @@protoc_insertion_point(class_scope:apsara.odps.cupid.protocol.RegisterTableReaderResponse)
  ))
_sym_db.RegisterMessage(RegisterTableReaderResponse)

RegisterTableWriterRequest = _reflection.GeneratedProtocolMessageType('RegisterTableWriterRequest', (_message.Message,), dict(
  DESCRIPTOR = _REGISTERTABLEWRITERREQUEST,
  __module__ = 'cupid_subprocess_service_pb2'
  # @@protoc_insertion_point(class_scope:apsara.odps.cupid.protocol.RegisterTableWriterRequest)
  ))
_sym_db.RegisterMessage(RegisterTableWriterRequest)

RegisterTableWriterResponse = _reflection.GeneratedProtocolMessageType('RegisterTableWriterResponse', (_message.Message,), dict(
  DESCRIPTOR = _REGISTERTABLEWRITERRESPONSE,
  __module__ = 'cupid_subprocess_service_pb2'
  # @@protoc_insertion_point(class_scope:apsara.odps.cupid.protocol.RegisterTableWriterResponse)
  ))
_sym_db.RegisterMessage(RegisterTableWriterResponse)

CommitActionInfo = _reflection.GeneratedProtocolMessageType('CommitActionInfo', (_message.Message,), dict(
  DESCRIPTOR = _COMMITACTIONINFO,
  __module__ = 'cupid_subprocess_service_pb2'
  # @@protoc_insertion_point(class_scope:apsara.odps.cupid.protocol.CommitActionInfo)
  ))
_sym_db.RegisterMessage(CommitActionInfo)

CommitTableFilesRequest = _reflection.GeneratedProtocolMessageType('CommitTableFilesRequest', (_message.Message,), dict(
  DESCRIPTOR = _COMMITTABLEFILESREQUEST,
  __module__ = 'cupid_subprocess_service_pb2'
  # @@protoc_insertion_point(class_scope:apsara.odps.cupid.protocol.CommitTableFilesRequest)
  ))
_sym_db.RegisterMessage(CommitTableFilesRequest)

CommitTableFilesResponse = _reflection.GeneratedProtocolMessageType('CommitTableFilesResponse', (_message.Message,), dict(
  DESCRIPTOR = _COMMITTABLEFILESRESPONSE,
  __module__ = 'cupid_subprocess_service_pb2'
  # @@protoc_insertion_point(class_scope:apsara.odps.cupid.protocol.CommitTableFilesResponse)
  ))
_sym_db.RegisterMessage(CommitTableFilesResponse)


DESCRIPTOR.has_options = True
DESCRIPTOR._options = _descriptor._ParseOptions(descriptor_pb2.FileOptions(), _b('B\033CupidSubProcessServiceProto\200\001\001\210\001\001\220\001\001'))

_CUPIDSUBPROCESSSERVICE = _descriptor.ServiceDescriptor(
  name='CupidSubProcessService',
  full_name='apsara.odps.cupid.protocol.CupidSubProcessService',
  file=DESCRIPTOR,
  index=0,
  options=None,
  serialized_start=1317,
  serialized_end=1966,
  methods=[
  _descriptor.MethodDescriptor(
    name='GetSplits',
    full_name='apsara.odps.cupid.protocol.CupidSubProcessService.GetSplits',
    index=0,
    containing_service=None,
    input_type=_GETSPLITSREQUEST,
    output_type=_GETSPLITSRESPONSE,
    options=None,
  ),
  _descriptor.MethodDescriptor(
    name='RegisterTableReader',
    full_name='apsara.odps.cupid.protocol.CupidSubProcessService.RegisterTableReader',
    index=1,
    containing_service=None,
    input_type=_REGISTERTABLEREADERREQUEST,
    output_type=_REGISTERTABLEREADERRESPONSE,
    options=None,
  ),
  _descriptor.MethodDescriptor(
    name='RegisterTableWriter',
    full_name='apsara.odps.cupid.protocol.CupidSubProcessService.RegisterTableWriter',
    index=2,
    containing_service=None,
    input_type=_REGISTERTABLEWRITERREQUEST,
    output_type=_REGISTERTABLEWRITERRESPONSE,
    options=None,
  ),
  _descriptor.MethodDescriptor(
    name='CommitTableFiles',
    full_name='apsara.odps.cupid.protocol.CupidSubProcessService.CommitTableFiles',
    index=3,
    containing_service=None,
    input_type=_COMMITTABLEFILESREQUEST,
    output_type=_COMMITTABLEFILESRESPONSE,
    options=None,
  ),
  _descriptor.MethodDescriptor(
    name='GetSplitsMeta',
    full_name='apsara.odps.cupid.protocol.CupidSubProcessService.GetSplitsMeta',
    index=4,
    containing_service=None,
    input_type=_GETSPLITSMETAREQUEST,
    output_type=_GETSPLITSMETARESPONSE,
    options=None,
  ),
])

CupidSubProcessService = service_reflection.GeneratedServiceType('CupidSubProcessService', (_service.Service,), dict(
  DESCRIPTOR = _CUPIDSUBPROCESSSERVICE,
  __module__ = 'cupid_subprocess_service_pb2'
  ))

CupidSubProcessService_Stub = service_reflection.GeneratedServiceStubType('CupidSubProcessService_Stub', (CupidSubProcessService,), dict(
  DESCRIPTOR = _CUPIDSUBPROCESSSERVICE,
  __module__ = 'cupid_subprocess_service_pb2'
  ))


# @@protoc_insertion_point(module_scope)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2022 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .core import LazyLoad
from .resource import Resource
from .. import serializers, utils


class Function(LazyLoad):
    """
    Function can be used in UDF when user writes a SQL.
    """
    __slots__ = '_resources_objects', '_owner_changed'

    _root = 'Function'

    name = serializers.XMLNodeField('Alias')
    _owner = serializers.XMLNodeField('Owner')
    creation_time = serializers.XMLNodeField('CreationTime', parse_callback=utils.parse_rfc822)
    class_type = serializers.XMLNodeField('ClassType')
    _resources = serializers.XMLNodesField('Resources', 'ResourceName')
    is_sql_function = serializers.XMLNodeField('IsSqlFunction')
    is_embedded_function = serializers.XMLNodeField('IsEmbeddedFunction')
    program_language = serializers.XMLNodeField('ProgramLanguage')
    code = serializers.XMLNodeField('Code')
    file_name = serializers.XMLNodeField('FileName')

    def __init__(self, **kwargs):
        self._resources_objects = None
        self._owner_changed = False

        resources = kwargs.pop('resources', None)
        if 'owner' in kwargs:
            kwargs['_owner'] = kwargs.pop('owner')
        super(Function, self).__init__(**kwargs)
        if resources is not None:
            self.resources = resources

    def reload(self):
        resp = self._client.get(
            self.resource(), curr_schema=self._get_schema_name()
        )
        self.parse(self._client, resp, obj=self)

    @property
    def resources(self):
        """
        Return all the resources which this function refer to.

        :return: resources
        :rtype: list

        .. seealso:: :class:`odps.models.Resource`
        """

        if self._resources_objects is not None:
            return self._resources_objects

        resources = self.parent.parent.resources
        resources = [resources[name] for name in self._resources or ()]
        self._resources_objects = resources
        return resources

    @resources.setter
    def resources(self, value):
        def get_resource_name(res):
            if isinstance(res, Resource):
                schema_name = res._get_schema_name()
                if res.project.name == self.project.name and schema_name is None:
                    return res.name
                return res.full_resource_name
            else:
                return res

        self._resources_objects = None
        self._resources = [get_resource_name(res) for res in value]

    @property
    def owner(self):
        return self._owner

    @owner.setter
    def owner(self, value):
        self._owner_changed = True
        self._owner = value

    def update(self):
        """
        Update this function.

        :return: None
        """
        if self._owner_changed:
            self.update_owner(self.owner)

        self._resources = [res.name for res in self.resources]
        return self.parent.update(self)

    def update_owner(self, new_owner):
        params = {}
        schema_name = self._get_schema_name()
        if schema_name:
            params['curr_schema'] = schema_name

        headers = {
            'x-odps-owner': new_owner
        }
        self._client.put(
            self.resource(), None, action="updateowner", params=params, headers=headers
        )

    def drop(self):
        """
        Delete this Function.

        :return: None
        """

        return self.parent.delete(self)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2022 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from sqlalchemy.ext.compiler import compiles

from ..sqlalchemy.ext import SACreateTempTableAs


class CreateTempTableAs(SACreateTempTableAs):

    def __init__(self, name, query, expr):
        super(CreateTempTableAs, self).__init__(name, query)
        self.expr = expr


@compiles(CreateTempTableAs)
def create_temp_table_as(element, compiler, **kw):
    return 'CREATE TEMP TABLE %s WITH(appendonly=true, orientation=heap, bucketnum=1) ' \
           'AS %s DISTRIBUTED BY (%s)' % (
        element.name,
        compiler.process(element.query),
        ', '.join(element.expr.schema.names)  # we just distribute by all columns
    )



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2022 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import random
import sys
import tarfile
import types
import warnings
import zipfile
from collections import defaultdict


_SEARCH_ORDER = [
    ('.py', False),
    ('/__init__.py', True),
]


try:
    os.path.exists('/tmp')
    ALLOW_BINARY = True
except:
    ALLOW_BINARY = False


if sys.version_info[0] <= 2:
    iterkeys = lambda d: d.iterkeys()
    itervalues = lambda d: d.itervalues()
    string_types = (basestring, unicode)
    BASE = object
else:
    import importlib.abc

    iterkeys = lambda d: d.keys()
    itervalues = lambda d: d.values()
    string_types = (str, bytes)
    BASE = importlib.abc.MetaPathFinder


def _clean_extract():
    if CompressImporter._extract_path:
        import shutil
        shutil.rmtree(CompressImporter._extract_path, ignore_errors=True)


class CompressImportError(ImportError):
    """Exception raised by CompressImporter objects."""
    pass


class CompressImporter(BASE):
    """
    A PEP-302-style importer that can import from a zipfile.

    Just insert or append this class (not an instance) to sys.path_hooks
    and you're in business.  Instances satisfy both the 'importer' and
    'loader' APIs specified in PEP 302.
    """
    _extract_path = None

    def __init__(self, *compressed_files, **kwargs):
        """
        Constructor.

        Args:
          compressed_files zipfile.ZipFile or tarfile.TarFile
        """
        self._files = []
        self._prefixes = defaultdict(lambda: set(['']))
        self._extract_binary = kwargs.get('extract_binary', kwargs.get('extract', False))
        self._extract_all = kwargs.get('extract_all', False)
        self._supersede = kwargs.get('supersede', True)

        for f in compressed_files:
            if isinstance(f, zipfile.ZipFile):
                bin_package = any(
                    n.endswith('.so') or n.endswith('.pxd') or n.endswith('.dylib')
                    for n in f.namelist()
                )
                need_extract = True
            elif isinstance(f, tarfile.TarFile):
                bin_package = any(
                    m.name.endswith('.so') or m.name.endswith('.pxd') or m.name.endswith('.dylib')
                    for m in f.getmembers()
                )
                need_extract = True
            elif isinstance(f, dict):
                bin_package = any(
                    name.endswith('.so') or name.endswith('.pxd') or name.endswith('.dylib')
                    for name in iterkeys(f)
                )
                need_extract = False
            elif isinstance(f, list):
                bin_package = any(
                    name.endswith('.so') or name.endswith('.pxd') or name.endswith('.dylib')
                    for name in f
                )
                need_extract = False
            else:
                raise TypeError('Compressed file can only be zipfile.ZipFile or tarfile.TarFile')

            if bin_package:
                # binary packages need to be extracted before use
                if not ALLOW_BINARY:
                    raise SystemError(
                        'Cannot load binary package. It is quite possible that you are using an old '
                        'MaxCompute service which does not support binary packages. If this is '
                        'not true, please set `odps.isolation.session.enable` to True or ask your '
                        'project owner to change project-level configuration.'
                    )
                if need_extract:
                    f = self._extract_archive(f)
            elif need_extract and self._extract_all:
                # when it is forced to extract even if it is a text package, also extract
                f = self._extract_archive(f)

            prefixes = set([''])
            dir_prefixes = set()  # only for lists or dicts
            if isinstance(f, zipfile.ZipFile):
                for name in f.namelist():
                    name = name if name.endswith('/') else (name.rsplit('/', 1)[0] + '/')
                    if name in prefixes:
                        continue
                    try:
                        f.getinfo(name + '__init__.py')
                    except KeyError:
                        prefixes.add(name)
            elif isinstance(f, tarfile.TarFile):
                for member in f.getmembers():
                    name = member.name if member.isdir() else member.name.rsplit('/', 1)[0]
                    if name in prefixes:
                        continue
                    try:
                        f.getmember(name + '/__init__.py')
                    except KeyError:
                        prefixes.add(name + '/')
            elif isinstance(f, (list, dict)):
                # Force ArchiveResource to run under binary mode to resolve manually
                # opening __file__ paths in pure-python code.
                if ALLOW_BINARY:
                    bin_package = True

                rendered_names = set()
                for name in f:
                    name = name.replace(os.sep, '/')
                    rendered_names.add(name)

                for name in rendered_names:
                    name = name if name.endswith('/') else (name.rsplit('/', 1)[0] + '/')
                    if name in prefixes or '/tests/' in name or '/__pycache__/' in name:
                        continue
                    if name + '__init__.py' not in rendered_names:
                        prefixes.add(name)
                        dir_prefixes.add(name)
                    else:
                        if '/' in name.rstrip('/'):
                            ppath = name.rstrip('/').rsplit('/', 1)[0]
                        else:
                            ppath = ''
                        prefixes.add(ppath)
                        dir_prefixes.add(ppath)

            # make sure only root packages are included, otherwise relative imports might be broken
            # NOTE that it is needed to check sys.path duplication after all pruning done,
            #  otherwise path might be error once CompressImporter is called twice.
            path_patch = []
            for p in sorted(dir_prefixes):
                parent_exist = False
                for pp in path_patch:
                    if p[:len(pp)] == pp:
                        parent_exist = True
                        break
                if parent_exist:
                    continue
                path_patch.append(p)

            if bin_package:
                path_patch = [p for p in path_patch if p not in sys.path]
                if self._supersede:
                    sys.path = path_patch + sys.path
                else:
                    sys.path = sys.path + path_patch

                try:
                    self._load_dynlibs(path_patch)
                except:
                    warnings.warn(
                        "Failed to call _load_dynlibs. Dynamic libraries might not be loaded properly."
                    )
            else:
                self._files.append(f)
                if path_patch:
                    path_patch = [p for p in path_patch if p not in sys.path]
                    self._prefixes[id(f)] = sorted([''] + path_patch)
                elif prefixes:
                    self._prefixes[id(f)] = sorted(prefixes)

    def _extract_archive(self, archive):
        if not self._extract_binary and not self._extract_all:
            raise SystemError(
                'We do not allow file-type resource for binary packages. Please upload an '
                'archive-typed resource instead.'
            )

        cls = type(self)
        if not cls._extract_path:
            import tempfile
            import atexit
            cls._extract_path = tempfile.mkdtemp(prefix='tmp-pyodps-')
            atexit.register(_clean_extract)

        extract_dir = os.path.join(cls._extract_path,
                                   'archive-' + str(random.randint(100000, 999999)))
        os.makedirs(extract_dir)
        if isinstance(archive, zipfile.ZipFile):
            archive.extractall(extract_dir)
        elif isinstance(archive, tarfile.TarFile):
            archive.extractall(extract_dir)

        mock_archive = list()
        for root, dirs, files in os.walk(extract_dir):
            for name in files:
                full_name = os.path.join(root, name)
                mock_archive.append(full_name)
        return mock_archive

    @classmethod
    def _load_dynlibs(cls, path_patch):
        import ctypes
        import glob

        # put dynlibs dir into LD_LIBRARY_PATH env
        library_paths = []
        for path in path_patch:
            dynlib_path = os.path.join(path, "dynlibs")
            if not os.path.isdir(dynlib_path):
                continue

            library_paths.append(dynlib_path)
            cwd = os.getcwd()
            try:
                os.chdir(dynlib_path)
                for so_file in glob.glob("*.so"):
                    try:
                        ctypes.cdll.LoadLibrary(so_file)
                    except:
                        warnings.warn(
                            "Failed to load dynamic library %s" % so_file, ImportWarning
                        )
            finally:
                os.chdir(cwd)
        if library_paths:
            if "LD_LIBRARY_PATH" in os.environ:
                os.environ["LD_LIBRARY_PATH"] = (
                    os.environ["LD_LIBRARY_PATH"] + ":" + ":".join(library_paths)
                )
            else:
                os.environ["LD_LIBRARY_PATH"] = ":".join(library_paths)

    def _get_info(self, fullmodname):
        """
        Internal helper for find_module() and load_module().

        Args:
          fullmodname: The dot-separated full module name, e.g. 'django.core.mail'.

        Returns:
          A tuple (submodname, is_package, relpath, fileobj) where:
            submodname: The final component of the module name, e.g. 'mail'.
            is_package: A bool indicating whether this is a package.
            relpath: The path to the module's source code within to the zipfile or tarfile.
            fileobj: The file object

        Raises:
          ImportError if the module is not found in the archive.
        """
        parts = fullmodname.split('.')
        submodname = parts[-1]
        for f in self._files:
            for prefix in self._prefixes[id(f)]:
                for suffix, is_package in _SEARCH_ORDER:
                    l = [prefix] + parts[:-1] + [submodname + suffix.replace('/', os.sep)]
                    relpath = os.path.join(*l)
                    try:
                        relpath = relpath.replace(os.sep, '/')
                        if isinstance(f, zipfile.ZipFile):
                            f.getinfo(relpath)
                        elif isinstance(f, tarfile.TarFile):
                            f.getmember(relpath)
                        else:
                            if relpath not in f:
                                raise KeyError
                    except KeyError:
                        pass
                    else:
                        return submodname, is_package, relpath, f

        msg = 'Can\'t find module %s' % fullmodname
        raise CompressImportError(msg)

    def _get_source(self, fullmodname):
        """
        Internal helper for load_module().

        Args:
          fullmodname: The dot-separated full module name, e.g. 'django.core.mail'.

        Returns:
          A tuple (submodname, is_package, fullpath, source) where:
            submodname: The final component of the module name, e.g. 'mail'.
            is_package: A bool indicating whether this is a package.
            fullpath: The path to the module's source code including the
              zipfile's or tarfile's filename.
            source: The module's source code.

        Raises:
          ImportError if the module is not found in the archive.
        """

        submodname, is_package, relpath, fileobj = self._get_info(fullmodname)
        fullpath = '%s%s%s' % (fileobj, os.sep, relpath)
        if isinstance(fileobj, zipfile.ZipFile):
            source = fileobj.read(relpath.replace(os.sep, '/'))
        elif isinstance(fileobj, tarfile.TarFile):
            source = fileobj.extractfile(relpath.replace(os.sep, '/')).read()
        elif isinstance(fileobj, dict):
            source = fileobj[relpath.replace(os.sep, '/')].read()
        else:
            source = open(fileobj[relpath.replace(os.sep, '/')], 'rb').read()
        source = source.replace(b'\r\n', b'\n').replace(b'\r', b'\n')
        return submodname, is_package, fullpath, source

    def find_module(self, fullmodname, path=None):
        """
        PEP-302-compliant find_module() method.

        Args:
          fullmodname: The dot-separated full module name, e.g. 'django.core.mail'.
          path: Optional and ignored; present for API compatibility only.

        Returns:
          None if the module isn't found in the archive; self if it is found.
        """
        try:
            self._get_info(fullmodname)
        except ImportError:
            return None
        else:
            return self

    def load_module(self, fullmodname):
        """
        PEP-302-compliant load_module() method.

        Args:
          fullmodname: The dot-separated full module name, e.g. 'django.core.mail'.

        Returns:
          The module object constructed from the source code.

        Raises:
          SyntaxError if the module's source code is syntactically incorrect.
          ImportError if there was a problem accessing the source code.
          Whatever else can be raised by executing the module's source code.
        """
        submodname, is_package, fullpath, source = self._get_source(fullmodname)
        code = compile(source, fullpath, 'exec')
        mod = sys.modules.get(fullmodname)
        try:
            if mod is None:
                mod = sys.modules[fullmodname] = types.ModuleType(fullmodname)
            mod.__loader__ = self
            mod.__file__ = fullpath
            mod.__name__ = fullmodname
            if is_package:
                mod.__path__ = [os.path.dirname(mod.__file__)]
            exec(code, mod.__dict__)
        except:
            if fullmodname in sys.modules:
                del sys.modules[fullmodname]
            raise
        return mod



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2022 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .expressions import Scalar, Expr, TableSchema, SequenceExpr, ListSequenceExpr, DictSequenceExpr, \
    ListScalar, DictScalar, Column
from .element import AnyOp, ElementWise
from .collections import RowAppliedCollectionExpr
from .. import types as df_types
from ..utils import to_collection
from . import utils


def _scalar(val, tp=None):
    if val is None:
        return
    if isinstance(val, Expr):
        return val
    if isinstance(val, (tuple, list)):
        return type(val)(_scalar(it, tp=tp) for it in val)
    else:
        return Scalar(_value=val, _value_type=tp)


def explode(expr, *args, **kwargs):
    """
    Expand list or dict data into multiple rows

    :param expr: list / dict sequence / scalar
    :return:
    """
    if not isinstance(expr, Column):
        expr = to_collection(expr)[expr.name]

    if isinstance(expr, SequenceExpr):
        dtype = expr.data_type
    else:
        dtype = expr.value_type

    func_name = 'EXPLODE'
    if args and isinstance(args[0], (list, tuple, set)):
        names = list(args[0])
    else:
        names = args
    pos = kwargs.get('pos', False)
    if isinstance(expr, ListSequenceExpr):
        if pos:
            func_name = 'POSEXPLODE'
            typos = [df_types.int64, dtype.value_type]
            if not names:
                names = [expr.name + '_pos', expr.name]
            if len(names) == 1:
                names = [names[0] + '_pos', names[0]]
            if len(names) != 2:
                raise ValueError("The length of parameter 'names' should be exactly 1.")
        else:
            typos = [dtype.value_type]
            if not names:
                names = [expr.name]
            if len(names) != 1:
                raise ValueError("The length of parameter 'names' should be exactly 1.")
    elif isinstance(expr, DictSequenceExpr):
        if pos:
            raise ValueError('Cannot support explosion with pos on dicts.')
        typos = [dtype.key_type, dtype.value_type]
        if not names:
            names = [expr.name + '_key', expr.name + '_value']
        if len(names) != 2:
            raise ValueError("The length of parameter 'names' should be exactly 2.")
    else:
        raise ValueError('Cannot explode expression with type %s' % type(expr).__name__)
    schema = TableSchema.from_lists(names, typos)
    return RowAppliedCollectionExpr(_input=expr.input, _func=func_name, _schema=schema,
                                    _fields=[expr], _keep_nulls=kwargs.get('keep_nulls', False))


def composite_op(expr, output_expr_cls, output_type=None, **kwargs):
    input_args = kwargs.copy()
    input_args['_input'] = expr

    def check_is_sequence(arg):
        a = input_args.get(arg)
        if isinstance(a, SequenceExpr):
            return True
        elif isinstance(a, (list, tuple)):
            return any(isinstance(la, SequenceExpr) for la in a)
        else:
            return False

    is_sequence = any(check_is_sequence(a) for a in output_expr_cls._args)

    if is_sequence:
        output_type = output_type or expr.data_type
        return output_expr_cls(_data_type=output_type, _input=expr, **kwargs)
    else:
        output_type = output_type or expr.value_type
        return output_expr_cls(_value_type=output_type, _input=expr, **kwargs)


class CompositeOp(ElementWise):
    def accept(self, visitor):
        visitor.visit_composite_op(self)


class CompositeBuilderOp(AnyOp):
    @property
    def node_name(self):
        return self.__class__.__name__

    def accept(self, visitor):
        visitor.visit_composite_op(self)


class ListDictLength(CompositeOp):
    __slots__ = ()


class ListDictGetItem(CompositeOp):
    _args = '_input', '_key', '_negative_handled'


class ListContains(CompositeOp):
    _args = '_input', '_value',


class ListSort(CompositeOp):
    __slots__ = ()


class DictKeys(CompositeOp):
    __slots__ = ()


class DictValues(CompositeOp):
    __slots__ = ()


def _scan_inputs(seq, dtype=None):
    if not seq:
        raise TypeError('Inputs should not be empty')

    seq = [_scalar(a) for a in seq]

    arg_types = set()
    arg_cat = set()
    for a in seq:
        if isinstance(a, SequenceExpr):
            arg_types.add(a.data_type)
            arg_cat.add(SequenceExpr)
        else:
            arg_types.add(a.value_type)
            arg_cat.add(Scalar)

    if dtype is not None:
        if not all(dtype.can_implicit_cast(t) for t in arg_types):
            raise TypeError('Not all given value can be implicitly casted')
    else:
        if len(arg_types) == 1:
            dtype = arg_types.pop()
            if isinstance(dtype, df_types.Integer):
                if dtype != df_types.int64 and df_types.int32.can_implicit_cast(dtype):
                    dtype = df_types.int32
            elif isinstance(dtype, df_types.Float):
                dtype = df_types.float64
        else:
            if all(df_types.int32.can_implicit_cast(t) for t in arg_types):
                dtype = df_types.int32
            elif all(df_types.int64.can_implicit_cast(t) for t in arg_types):
                dtype = df_types.int64
            elif all(df_types.float64.can_implicit_cast(t) for t in arg_types):
                dtype = df_types.float64
            else:
                raise TypeError('Types of inputs should be the same')
    if SequenceExpr in arg_cat:
        return seq, '_data_type', dtype
    else:
        return seq, '_value_type', dtype


class ListBuilder(CompositeBuilderOp):
    _args = '_values',


class DictBuilder(CompositeBuilderOp):
    _args = '_keys', '_values'


def _len(expr):
    """
    Retrieve length of a list or dict sequence / scalar.

    :param expr: list or dict sequence / scalar
    :return:
    """
    return composite_op(expr, ListDictLength, df_types.int64)


def _getitem(expr, key):
    if isinstance(expr, SequenceExpr):
        dtype = expr.data_type.value_type
    else:
        dtype = expr.value_type.value_type
    return composite_op(expr, ListDictGetItem, dtype, _key=_scalar(key))


def _sort(expr):
    """
    Retrieve sorted list

    :param expr: list sequence / scalar
    :return:
    """
    return composite_op(expr, ListSort)


def _contains(expr, value):
    """
    Check whether certain value is in the inspected list

    :param expr: list sequence / scalar
    :param value: value to inspect
    :return:
    """
    return composite_op(expr, ListContains, df_types.boolean, _value=_scalar(value))


def _keys(expr):
    """
    Retrieve keys of a dict

    :param expr: dict sequence / scalar
    :return:
    """
    if isinstance(expr, SequenceExpr):
        dtype = expr.data_type
    else:
        dtype = expr.value_type
    return composite_op(expr, DictKeys, df_types.List(dtype.key_type))


def _values(expr):
    """
    Retrieve values of a dict

    :param expr: dict sequence / scalar
    :return:
    """
    if isinstance(expr, SequenceExpr):
        dtype = expr.data_type
    else:
        dtype = expr.value_type
    return composite_op(expr, DictValues, df_types.List(dtype.value_type))


def make_list(*args, **kwargs):
    dtype = kwargs.get('type')
    if dtype is not None:
        dtype = df_types.validate_data_type(dtype)

    kwargs = dict()
    kwargs['_values'], k, typ = _scan_inputs(args, dtype)
    kwargs[k] = df_types.List(typ)

    return ListBuilder(**kwargs)


def make_dict(*args, **kwargs):
    if len(args) % 2 != 0:
        raise ValueError('Num of inputs to build a dict should be even')

    key_type = kwargs.get('key_type')
    if key_type is not None:
        key_type = df_types.validate_data_type(key_type)

    value_type = kwargs.get('value_type')
    if value_type is not None:
        value_type = df_types.validate_data_type(value_type)

    kwargs = dict()
    keys = list(args[0::2])
    values = list(args[1::2])

    kwargs['_keys'], k1, key_type = _scan_inputs(keys, key_type)
    kwargs['_values'], k2, value_type = _scan_inputs(values, value_type)

    k = '_data_type' if '_data_type' in (k1, k2) else '_value_type'
    kwargs[k] = df_types.Dict(key_type, value_type)

    return DictBuilder(**kwargs)


_list_methods = dict(
    __getitem__=_getitem,
    len=_len,
    sort=_sort,
    contains=_contains,
    explode=explode,
)

_dict_methods = dict(
    __getitem__=_getitem,
    len=_len,
    keys=_keys,
    values=_values,
    explode=explode,
)

utils.add_method(ListSequenceExpr, _list_methods)
utils.add_method(ListScalar, _list_methods)
utils.add_method(DictSequenceExpr, _dict_methods)
utils.add_method(DictScalar, _dict_methods)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2022 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import contextlib

import requests

from . import cache_parent
from .. import serializers
from ..compat import Enum
from ..errors import OSSSignUrlError
from .volume_fs import FSVolumeObject, FSVolumeObjects, FSVolume


class SignUrlMethod(Enum):
    GET = "get"
    PUT = "put"


class ExternalVolumeObject(FSVolumeObject):
    @classmethod
    def _get_base_cls(cls):
        return ExternalVolumeObject

    @classmethod
    def _get_file_cls(cls):
        return ExternalVolumeFile

    @classmethod
    def _get_dir_cls(cls):
        return ExternalVolumeDir

    def get_sign_url(self, method, seconds=None):
        if isinstance(method, SignUrlMethod):
            method = method.value
        params = {'sign_url': method.lower()}
        if seconds:
            params["expire_seconds"] = seconds
        headers = {'x-odps-volume-fs-path': self.path}

        schema_name = self.volume._get_schema_name()
        if schema_name is not None:
            params["curr_schema"] = schema_name
        resp = self._client.get(
            self.parent.resource(), action='meta', params=params, headers=headers
        )
        self.parse(self._client, resp, obj=self)
        return self.sign_url

    def _request_sign_url(self, path, method, *args, **kw):
        if path:
            path = self.path.rstrip('/') + '/' + path.lstrip('/')
        else:
            path = self.path
        vol_rel_path = path[len(self.volume.name) + 1:]
        sign_url = self.volume.get_sign_url(vol_rel_path, method)
        if method == SignUrlMethod.PUT:
            resp = requests.put(sign_url, *args, **kw)
        else:
            resp = requests.get(sign_url, *args, **kw)
        resp.volume_path = path
        self._check_response(resp)
        return resp

    @staticmethod
    def _check_response(resp):
        # when response code is a string, just skip
        if hasattr(resp, "status_code") and resp.status_code >= 400:
            try:
                import oss2.exceptions

                oss_exc = oss2.exceptions.make_exception(resp.raw)
                raise OSSSignUrlError(oss_exc)
            except ImportError:
                raise OSSSignUrlError(resp.content)

    def _delete(self, recursive=False):
        """
        Delete current directory.

        :param recursive: indicate whether a recursive deletion should be performed.
        """
        params = {'recursive': str(recursive).lower()}
        headers = {'x-odps-volume-fs-path': self.path.rstrip("/")}
        self._del_cache(self.path)
        self._client.delete(
            self.parent.resource(),
            params=params,
            headers=headers,
            curr_schema=self.parent._get_schema_name(),
        )

    @contextlib.contextmanager
    def _open_reader(self, path):
        """
        Open a volume file and read contents in it.

        :param str path: file name to be opened
        :return: file reader

        :Example:
        >>> with fs_dir.open_reader('file') as reader:
        >>>     [print(line) for line in reader]
        """
        req = self._request_sign_url(path, SignUrlMethod.GET, stream=True)
        yield req.raw

    @contextlib.contextmanager
    def _open_writer(self, path, **kwargs):
        """
        Open a volume file and write contents into it.

        :param str path: file name to be opened
        :return: file reader

        :Example:
        >>> with fs_dir.open_writer('file') as reader:
        >>>     writer.write('some content')
        """
        from ..tunnel.io import RequestsIO

        if kwargs.pop("replication", None) is not None:  # pragma: no cover
            raise TypeError("External volume does not support replication argument")

        def put_func(data):
            self._request_sign_url(path, SignUrlMethod.PUT, data=data)

        rio = RequestsIO(put_func)
        try:
            rio.start()
            yield rio
        finally:
            rio.finish()


@cache_parent
class ExternalVolumeFile(ExternalVolumeObject):
    def __init__(self, **kw):
        super(ExternalVolumeFile, self).__init__(**kw)
        self._isdir = False

    def delete(self):
        """
        Delete current file.
        """
        return self._delete(False)

    def open_reader(self):
        """
        Open current file and read contents in it.
        :return: file reader

        :Example:
        >>> with fs_file.open_reader('file') as reader:
        >>>     [print(line) for line in reader]
        """
        return self._open_reader(None)

    def open_writer(self, **kw):
        return self._open_writer(None, **kw)


@cache_parent
class ExternalVolumeDir(ExternalVolumeObject):
    def __init__(self, **kw):
        super(ExternalVolumeDir, self).__init__(**kw)
        self._isdir = True

    @property
    def objects(self):
        return ExternalVolumeObjects(parent=self, client=self._client)

    def create_dir(self, path):
        """
        Creates and returns a sub-directory under the current directory.
        :param str path: directory name to be created
        :return: directory object
        :rtype: :class:`odps.models.FSVolumeDir`
        """
        path = path.strip("/") + "/"
        resp = self._request_sign_url(path, SignUrlMethod.PUT, b"")
        dir_object = type(self)(path=resp.volume_path, parent=self.parent, client=self._client)
        dir_object.reload()
        return dir_object

    def __contains__(self, item):
        return item in self.objects

    def __iter__(self):
        return self.objects.iterate()

    def __getitem__(self, item):
        return self.objects[item]

    def delete(self, recursive=False):
        """
        Delete current directory.

        :param recursive: indicate whether a recursive deletion should be performed.
        """
        self._delete(recursive=recursive)

    def open_reader(self, path):
        """
        Open a volume file and read contents in it.

        :param str path: file name to be opened
        :return: file reader

        :Example:
        >>> with fs_dir.open_reader('file') as reader:
        >>>     [print(line) for line in reader]
        """
        return self._open_reader(path)

    def open_writer(self, path, **kwargs):
        """
        Open a volume file and write contents into it.

        :param str path: file name to be opened
        :return: file reader

        :Example:
        >>> with fs_dir.open_writer('file') as reader:
        >>>     writer.write('some content')
        """
        return self._open_writer(path, **kwargs)


class ExternalVolumeObjects(FSVolumeObjects):
    objects = serializers.XMLNodesReferencesField(ExternalVolumeObject, 'Item')

    @classmethod
    def _get_single_object_cls(cls):
        return ExternalVolumeObject


class ExternalVolume(FSVolume):
    _dir_cls = ExternalVolumeDir

    def get_sign_url(self, path, method, seconds=None):
        path = "/" + self.name + "/" + path.lstrip("/")
        vol_file = ExternalVolumeFile(path=path, parent=self, client=self._client)
        return vol_file.get_sign_url(method, seconds=seconds)



# encoding: utf-8
# Copyright 1999-2022 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy

from ...compat import six, reduce
from ...df.expr.collections import CollectionExpr
from ..utils import MLField, FieldRole, FieldContinuity


"""
Base Operation
"""


class DFOperation(object):
    def execute(self, sources, target):
        """
        :type sources: list[DFAdapter]
        :type target: DFAdapter
        """
        fields = self._get_fields_list_from_eps(sources)
        target._ml_fields = copy.deepcopy(fields[0])

    @staticmethod
    def _get_fields_list_from_eps(dfs):
        def _filter_ml_fields(df):
            if isinstance(df, CollectionExpr):
                fields = set(c.name for c in df.schema.columns)
            else:
                fields = set([df.name])
            return [f for f in df._ml_fields if f.name in fields]

        return copy.deepcopy([_filter_ml_fields(ep) for ep in dfs])

    @staticmethod
    def _norm_name_set(name_set):
        if isinstance(name_set, six.string_types):
            return set(v.strip() for v in name_set.split(','))
        elif isinstance(name_set, (list, tuple)):
            return set(name_set)
        else:
            return name_set

    @staticmethod
    def _set_singleton_role(fields, role_mapping):
        roles = set(six.itervalues(role_mapping))
        for f in fields:
            if f.name in role_mapping:
                yield f.copy(role_mapping[f.name])
            elif roles & f.role:
                ret_field = f.copy()
                ret_field.role -= roles
                yield ret_field
            else:
                yield copy.deepcopy(f)
        for fname in set(six.iterkeys(role_mapping)) - set(f.name for f in fields):
            yield MLField(fname, 'EXPECTED', role_mapping[fname])

    @classmethod
    def _remove_field_roles(cls, fields, name_set, role):
        name_set = cls._norm_name_set(name_set)
        for f in fields:
            if f.name in name_set:
                ret_field = f.copy()
                ret_field.role -= set([role, ])
                yield ret_field
            else:
                yield copy.deepcopy(f)
        # if name does not appear in existing fields, we assume that it will be generated
        for name in name_set - set(f.name for f in fields):
            yield MLField(name, 'EXPECTED', role)

    @classmethod
    def _clear_field_roles(cls, fields, name_set):
        name_set = cls._norm_name_set(name_set)
        for f in fields:
            if f.name in name_set:
                ret_field = f.copy()
                ret_field.role = set()
                yield ret_field
            else:
                yield copy.deepcopy(f)
        # if name does not appear in existing fields, we assume that it will be generated
        for name in name_set - set(f.name for f in fields):
            yield MLField(name, 'EXPECTED', None)

    @classmethod
    def _add_field_roles(cls, fields, name_set, role, augment):
        name_set = cls._norm_name_set(name_set)
        for f in fields:
            if f.name in name_set:
                yield f.copy(role)
            elif role in f.role:
                if augment:
                    yield f.copy()
                else:
                    ret_field = f.copy()
                    ret_field.role.remove(role)
                    yield ret_field
            else:
                yield copy.deepcopy(f)
        # if name does not appear in existing fields, we assume that it will be generated
        for name in name_set - set(f.name for f in fields):
            yield MLField(name, 'EXPECTED', role)


"""
Attribution Operations
"""


class BatchRoleOperation(DFOperation):
    def __init__(self, fields, role, augment):
        self.feature_names = fields
        self.augment = augment
        self.role = role

    def execute(self, sources, target):
        """
        :type sources: list[DFAdapter]
        :type target: DFAdapter
        """
        fields = self._get_fields_list_from_eps(sources)
        feature_set = set(self.feature_names)
        target._ml_fields = list(self._add_field_roles(fields[0], feature_set, self.role, self.augment))


class ExcludeFieldsOperation(DFOperation):
    def __init__(self, fields):
        self.exclude_names = fields

    def execute(self, sources, target):
        """
        :type sources: list[DFAdapter]
        :type target: DFAdapter
        """
        fields = self._get_fields_list_from_eps(sources)
        feature_set = set(self.exclude_names)
        target._ml_fields = list(self._clear_field_roles(fields[0], feature_set))


class SingletonRoleOperation(DFOperation):
    def __init__(self, field_mapping, clear_feature=False):
        self.field_mapping = field_mapping
        self.clear_feature = clear_feature

    def execute(self, sources, target):
        """
        :type sources: list[DFAdapter]
        :type target: DFAdapter
        """
        fields = self._get_fields_list_from_eps(sources)
        ret_fields = fields[0]
        if self.clear_feature:
            ret_fields = list(self._remove_field_roles(ret_fields, set(six.iterkeys(self.field_mapping)),
                                                       FieldRole.FEATURE))
        target._ml_fields = list(self._set_singleton_role(ret_fields, self.field_mapping))


class FieldContinuityOperation(DFOperation):
    def __init__(self, continuity):
        self.continuity = dict((k, FieldContinuity.CONTINUOUS if v else FieldContinuity.DISCRETE)
                               for k, v in six.iteritems(continuity))

    def execute(self, sources, target):
        """
        :type sources: list[DFAdapter]
        :type target: DFAdapter
        """
        fields = self._get_fields_list_from_eps(sources)
        for f in fields[0]:
            if f.name in self.continuity:
                f.continuity = self.continuity[f.name]
        for field_name in set(six.iterkeys(self.continuity)) - set(f.name for f in fields[0]):
            fields[0].append(MLField(field_name, 'EXPECTED', None, continuity=self.continuity[field_name]))
        target._ml_fields = fields[0]


class FieldKVConfigOperation(DFOperation):
    def __init__(self, kv_config):
        self.kv_config = kv_config

    def execute(self, sources, target):
        """
        :type sources: list[DFAdapter]
        :type target: DFAdapter
        """
        fields = self._get_fields_list_from_eps(sources)
        for f in fields[0]:
            if f.name in self.kv_config:
                f.kv_config = self.kv_config[f.name]
        for field_name in set(six.iterkeys(self.kv_config)) - set(f.name for f in fields[0]):
            fields[0].append(MLField(field_name, 'EXPECTED', None, kv_config=self.kv_config[field_name]))
        target._ml_fields = fields[0]


"""
Modify Operations
"""


class StaticFieldChangeOperation(DFOperation):
    def __init__(self, fields, is_append=False):
        self.fields = [fields, ] if isinstance(fields, MLField) else fields
        self.is_append = is_append

    def execute(self, sources, target):
        """
        :type sources: list[DFAdapter]
        :type target: DFAdapter
        """
        existing_ml_fields = self._get_fields_list_from_eps(sources)
        dup_ml_fields = [f.copy() for f in self.fields]
        if self.is_append:
            src = list([copy.deepcopy(f) for f in existing_ml_fields[0]])
            src.extend(dup_ml_fields)
            target._ml_fields = src
        else:
            target._ml_fields = self.fields


class ProgrammaticFieldChangeOperation(DFOperation):
    def __init__(self, evaluator, is_append=False):
        self.is_append = is_append
        self.evaluator = evaluator

    def execute(self, sources, target):
        """
        :type sources: list[DFAdapter]
        :type target: DFAdapter
        """
        existing_ml_fields = self._get_fields_list_from_eps(sources)

        def gen_field(field_str):
            parts = field_str.split(':')
            if len(parts) > 2:
                roles = set(MLField.translate_role_name(rn) for rn in parts[2].strip().split('#')) - set([None])
            else:
                roles = set([FieldRole.FEATURE, ])
            return MLField(parts[0].strip(), parts[1].strip(), roles)

        fields = [gen_field(fstr) for fstr in self.evaluator().split(',')
                  if fstr.strip() != '']

        dup_ml_fields = [f.copy(None) for f in fields]
        if self.is_append:
            src = list([copy.deepcopy(f) for f in existing_ml_fields[0]])
            src.extend(dup_ml_fields)
            target._ml_fields = src
        else:
            target._ml_fields = fields


class MergeFieldsOperation(DFOperation):
    def __init__(self, auto_rename, selected_cols, excluded_cols):
        self._auto_rename = auto_rename
        self._selected_cols = selected_cols
        self._excluded_cols = excluded_cols

    def execute(self, sources, target):
        """
        :type sources: list[DFAdapter]
        :type target: DFAdapter
        """
        existing_fields = self._get_fields_list_from_eps(sources)

        def trans_columns(tid, fields):
            sel_cols = self._selected_cols.get(tid)
            exc_cols = self._excluded_cols.get(tid)
            for field in fields:
                if field.role is None:
                    continue
                if sel_cols and field.name not in sel_cols:
                    continue
                if exc_cols and field.name in exc_cols:
                    continue
                new_col = field.copy(field.role)
                if self._auto_rename:
                    new_col.name = 't%d_%s' % (tid, field.name)
                yield new_col

        new_fields = map(lambda tp: list(trans_columns(tp[0], tp[1])), enumerate(existing_fields))
        target._ml_fields = list(reduce(lambda fa, fb: fa + fb, new_fields, []))
        target._ml_uplink = sources
